import re

from scrubadub.detectors.catalogue import register_detector
from .base import RegexDetector
from ..filth import TwitterFilth


@register_detector
class TwitterDetector(RegexDetector):
    """Use regular expression magic to remove twitter usernames from dirty
    dirty ``text``.
    """
    filth_cls = TwitterFilth
    name = 'twitter'
    autoload = True

    # https://help.twitter.com/en/managing-your-account/twitter-username-rules#error
    # Twitter user names must be 15 or less charachtors and only contain a-zA-Z0-9_
    # Twitter and admin are not allowed in user names
    # (?<!\w) prevents it matching email addresses
    regex = re.compile((
        r"(?<!\w)@((?!((admin)|(twitter)))[a-z0-9_]){2,15}\b"
    ), re.VERBOSE | re.IGNORECASE)



import re
import warnings
from typing import Optional, ClassVar, Type, Generator, Pattern, Dict, Sequence

from ..filth import Filth
from ..import utils


class Detector(object):
    """This is the base class for all detectors.

    A simple example of how to make a new detector is given below:

    .. code:: pycon

        >>> import scrubadub
        >>> class MyFilth(scrubadub.filth.Filth):
        ...     type = 'mine'
        >>> class MyDetector(scrubadub.detectors.Detector):
        ...     name = 'my_fr_detector'
        ...     def iter_filth(self, text, document_name=None):
        ...         # This detector always returns this same Filth no matter the input.
        ...         # You should implement something better here.
        ...         yield MyFilth(beg=0, end=8, text='My stuff', document_name=document_name, detector_name=self.name)
        >>> scrubber = scrubadub.Scrubber()
        >>> scrubber.add_detector(MyDetector)
        >>> text = "My stuff can be found there."
        >>> scrubber.clean(text)
        '{{MINE}} can be found there.'

    You can also advertise a ``Detector`` as supporting a certain locale by defining the
    ```Detector.supported_local()``` function.
    """

    filth_cls = Filth  # type: ClassVar[Type[Filth]]
    name = 'detector'  # type: str
    autoload = False  # type: bool

    def __init__(self, name: Optional[str] = None, locale: str = 'en_US'):
        """Initialise the ``Detector``.

        :param name: Overrides the default name of the :class:``Detector``
        :type name: str, optional
        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str, optional
        """
        if getattr(self, 'name', 'detector') == 'detector' and getattr(self, 'filth_cls', None) is not None:
            if getattr(self.filth_cls, 'type', None) is not None and type(self) != Detector:
                self.name = self.filth_cls.type
                warnings.warn(
                    "Setting the detector name from the filth_cls.type is depreciated, please declare an explicit name"
                    "attribute on the class.",
                    DeprecationWarning
                )
        if name is not None:
            self.name = name

        self.locale = locale
        self.language, self.region = self.locale_split(locale)

        if hasattr(self, 'supported_locale'):
            if not self.supported_locale(locale=locale):  # type: ignore
                warnings.warn("Detector {} does not support the locale '{}'.".format(self.name, locale))

    locale_transform = staticmethod(utils.locale_transform)
    locale_split = staticmethod(utils.locale_split)

    def iter_filth(self, text: str, document_name: Optional[str] = None) -> Generator[Filth, None, None]:
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """
        raise NotImplementedError('must be implemented in derived classes')

    def iter_filth_documents(self, document_list: Sequence[str],
                             document_names: Sequence[Optional[str]]) -> Generator[Filth, None, None]:
        """Yields discovered filth in a list of documents.

        :param document_list: A list of documents to clean.
        :type document_list: List[str]
        :param document_names: A list containing the name of each document.
        :type document_names: List[str]
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """
        raise NotImplementedError('must be implemented in derived classes')


class RegexDetector(Detector):
    """Base class to match PII with a regex.

    This class requires that the ``filth_cls`` attribute be set to the class of the ``Filth`` that should be
    returned by this ``Detector``.

    .. code:: pycon

        >>> import re, scrubadub
        >>> class NewUrlDetector(scrubadub.detectors.RegexDetector):
        ...     name = 'new_url_detector'
        ...     filth_cls = scrubadub.filth.url.UrlFilth
        ...     regex = re.compile(r'https.*$', re.IGNORECASE)
        >>> scrubber = scrubadub.Scrubber(detector_list=[NewUrlDetector()])
        >>> text = u"This url will be found https://example.com"
        >>> scrubber.clean(text)
        'This url will be found {{URL}}'
    """

    regex = None  # type: Optional[Pattern[str]]
    filth_cls = Filth  # type: ClassVar[Type[Filth]]

    def iter_filth(self, text: str, document_name: Optional[str] = None) -> Generator[Filth, None, None]:
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """
        if not issubclass(self.filth_cls, Filth):
            raise TypeError(
                'filth_cls attribute of {} needs to be set to a subclass of the Filth class.'.format(self.__class__)
            )

        # Allow the regex to be in the detector as well  as the filth class
        if self.regex is None:
            warnings.warn('regex should be defined in the Detector and not in the Filth class', DeprecationWarning)
            if self.filth_cls.regex is not None:
                self.regex = self.filth_cls.regex

        if self.regex is None:
            raise ValueError('No regular expression has been specified for {}.'.format(self.__class__))

        for match in self.regex.finditer(text):
            yield self.filth_cls(match=match, detector_name=self.name, document_name=document_name,
                                 locale=self.locale)


class RegionLocalisedRegexDetector(RegexDetector):
    """Detector to detect ``Filth`` localised using regular expressions localised by the region"""
    region_regex = {}  # type: Dict[str, Pattern]

    def __init__(self, **kwargs):
        """Initialise the ``Detector``.

        :param name: Overrides the default name of the :class:``Detector``
        :type name: str, optional
        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str, optional
        """
        super(RegionLocalisedRegexDetector, self).__init__(**kwargs)

        # This will never be matched to anything.
        # It says anything where the next char is not an "a" character, but that is an "a".
        # This so the detector wont return filth if it doesn't have the region's correct regex.
        self.regex = re.compile(r'(?!a)a')

        if self.region in self.region_regex:
            self.regex = self.region_regex[self.region]

    @classmethod
    def supported_locale(cls, locale: str) -> bool:
        """Returns true if this ``Detector`` supports the given locale.

        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str
        :return: ``True`` if the locale is supported, otherwise ``False``
        :rtype: bool
        """
        language, region = cls.locale_split(locale)
        return region in cls.region_regex.keys()



import re
import phonenumbers

from faker import Faker
from typing import List

from .base import Filth
from .. import utils


class PhoneFilth(Filth):
    type = 'phone'

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        phone_number = ''
        language, region = utils.locale_split(faker._locales[0])
        results = []  # type: List[phonenumbers.PhoneNumberMatch]
        # Here I'm filtering for numbers that pass validation by the phonenumbers package
        while len(results) < 1:
            # Faker generates random numbers of the right format eg (###)###-####
            phone_number = re.sub(r'x.*$', '', faker.phone_number())
            # phonenumbers checks that they follow the rules around area codes and that they are possibly valid
            results = list(phonenumbers.PhoneNumberMatcher(phone_number, region))
        return phone_number



import re

from scrubadub.detectors.catalogue import register_detector
from .base import RegionLocalisedRegexDetector
from ..filth.postalcode import PostalCodeFilth


@register_detector
class PostalCodeDetector(RegionLocalisedRegexDetector):
    """Detects postal codes, currently only British post codes are supported."""
    filth_cls = PostalCodeFilth
    name = 'postalcode'
    autoload = True
    region_regex = {
        # Informed by https://en.wikipedia.org/wiki/Postcodes_in_the_United_Kingdom#Validation
        # and validated against https://osdatahub.os.uk/downloads/open/CodePointOpen
        'GB': re.compile(r"""
            (
                # Girobank postcode
                (?:[gG][iI][rR] {0,}0[aA]{2})|
                (?:  # British Overseas Territories in usual format
                    (?:
                        [aA][sS][cC][nN]|
                        [sS][tT][hH][lL]|
                        [tT][dD][cC][uU]|
                        [bB][bB][nN][dD]|
                        [bB][iI][qQ][qQ]|
                        [fF][iI][qQ][qQ]|
                        [pP][cC][rR][nN]|
                        [sS][iI][qQ][qQ]|
                        [iT][kK][cC][aA]
                    )
                    \ {0,}1[zZ]{2}
                )|
                (?:  # British Overseas Territories in zip-code format
                    (KY[0-9]|MSR|VG|AI)[ -]{0,}[0-9]{4}
                )|
                # (?:  # Bermuda including this causes too many false positives, so excluded for now
                #     [a-zA-Z]{2}\ {0,}[0-9]{2}
                # )|
                (?:  # British Forces Post Office
                    [Bb][Ff][Pp][Oo]\ {0,}[0-9]{1,4}
                )|
                (?:  # Mainland British postcodes
                    (?:
                        (?:[Ww][Cc][0-9][abehmnprvwxyABEHMNPRVWXY])|
                        (?:[Ee][Cc][1-4][abehmnprvwxyABEHMNPRVWXY])|
                        (?:[Nn][Ww]1[Ww])|
                        (?:[Ss][Ee]1[Pp])|
                        (?:[Ss][Ww]1[abehmnprvwxyABEHMNPRVWXY])|
                        (?:[EeNnWw]1[a-hjkpstuwA-HJKPSTUW])|
                        (?:[BbEeGgLlMmNnSsWw][0-9][0-9]?)|
                        (?:[a-pr-uwyzA-PR-UWYZ][a-hk-yxA-HK-XY][0-9][0-9]?)
                    )
                    \ {0,}[0-9][abd-hjlnp-uw-zABD-HJLNP-UW-Z]{2}
                )
            )
        """, re.VERBOSE),
    }



import inspect
import catalogue

from typing import Type, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from scrubadub.post_processors import PostProcessor

post_processor_catalogue = catalogue.create('scrubadub', 'post_processors', entry_points=True)


def register_post_processor(post_processor: Type['PostProcessor'], autoload: Optional[bool] = None,
                            index: Optional[int] = None) -> None:
    """Register a PostProcessor for use with the ``Scrubber`` class.

    You can use ``register_post_processor(NewPostProcessor)`` after your post-processor definition to automatically
    register it with the ``Scrubber`` class so that it can be used to process Filth.

    The argument ``autoload`` sets if a new ``Scrubber()`` instance should load this ``PostProcessor`` by default.

    :param post_processor: The ``PostProcessor`` to register with the scrubadub post-processor configuration.
    :type post_processor: PostProcessor class
    :param autoload: Whether to automatically load this ``Detector`` on ``Scrubber`` initialisation.
    :type autoload: bool
    :param index: The location/index in which this ``PostProcessor`` should be added.
    :type index: int
    """
    if not inspect.isclass(post_processor):
        raise ValueError("post_processor should be a class, not an instance.")

    if autoload is not None:
        post_processor.autoload = autoload

    if index is not None:
        post_processor.index = index

    post_processor_catalogue.register(post_processor.name, func=post_processor)


def remove_post_processor(post_processor: Union[Type['PostProcessor'], str]) -> None:
    """Remove an already registered post-processor.

    :param post_processor: The ``PostProcessor`` to register with the scrubadub post-processor configuration.
    :type post_processor: Union[Type['PostProcessor'], str]
    """
    if isinstance(post_processor, str):
        if post_processor in post_processor_catalogue:
            catalogue._remove((*post_processor_catalogue.namespace, post_processor))

    elif inspect.isclass(post_processor):
        if post_processor.name in post_processor_catalogue:
            catalogue._remove((*post_processor_catalogue.namespace, post_processor.name))

    else:
        raise ValueError("post-processor should be a class (not an instance) or a string.")



import re

from scrubadub.detectors.catalogue import register_detector
from scrubadub.detectors.base import RegionLocalisedRegexDetector
from scrubadub.filth import NationalInsuranceNumberFilth


@register_detector
class NationalInsuranceNumberDetector(RegionLocalisedRegexDetector):
    """Use regular expressions to remove the GB National Insurance number (NINO),
    Simple pattern matching, no checksum solution.
    """
    name = 'national_insurance_number'
    autoload = True
    filth_cls = NationalInsuranceNumberFilth
    # this regex is looking for NINO that does not begin with certain letters
    region_regex = {
        'GB': re.compile(
            r'(?!BG)(?!GB)(?!NK)(?!KN)(?!TN)(?!NT)(?!ZZ)(?:[A-CEGHJ-PR-TW-Z][A-CEGHJ-NPR-TW-Z])(?:\s*\d\s*){6}[A-D]',
            re.IGNORECASE | re.VERBOSE
        ),
    }



from faker import Faker

from .base import Filth


class LocationFilth(Filth):
    type = 'location'

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        return faker.city()



import warnings
from typing import Optional, Sequence, Generator, Dict, Type, Union, List

from . import detectors
from . import post_processors
from .detectors import Detector
from .post_processors import PostProcessor
from .filth import Filth


class Scrubber:
    """The Scrubber class is used to clean personal information out of dirty
    dirty text. It manages a set of ``Detector``'s that are each responsible
    for identifying ``Filth``. ``PostProcessor`` objects are used to alter
    the found Filth. This could be to replace the Filth with a hash or token.
    """

    def __init__(self, detector_list: Optional[Sequence[Union[Type[Detector], Detector, str]]] = None,
                 post_processor_list: Optional[Sequence[Union[Type[PostProcessor], PostProcessor, str]]] = None,
                 locale: Optional[str] = None):
        """Create a ``Scrubber`` object.

        :param detector_list: The list of detectors to use in this scrubber.
        :type detector_list: Optional[Sequence[Union[Type[Detector], Detector, str]]]
        :param post_processor_list: The locale that the phone number should adhere to.
        :type post_processor_list: Optional[Sequence[Union[Type[Detector], Detector, str]]]
        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str, optional
        """
        super().__init__()

        # instantiate all of the detectors which, by default, uses all of the
        # detectors that are in the detectors.types dictionary
        self._detectors = {}  # type: Dict[str, Detector]
        self._post_processors = []  # type: List[PostProcessor]

        if locale is None:
            locale = 'en_US'
        self._locale = locale  # type: str

        if detector_list is None:
            # First we gather all detectors that should automatically load
            detector_list = [
                detector
                for detector in detectors.catalogue.detector_catalogue.get_all().values()
                if detector.autoload and (
                    # Then we filter out ones that don't support the current locale
                    not hasattr(detector, 'supported_locale') or (
                        hasattr(detector, 'supported_locale') and
                        detector.supported_locale(locale)  # type: ignore
                    )
                )
            ]

        for detector in detector_list:
            self.add_detector(detector, warn=True)

        if post_processor_list is None:
            post_processor_list = [
                post_processor
                for post_processor in sorted(
                    post_processors.catalogue.post_processor_catalogue.get_all().values(),
                    key=lambda pp: pp.index,
                )
                if post_processor.autoload
            ]

        for post_processor in post_processor_list:
            self.add_post_processor(post_processor)

    def add_detector(self, detector: Union[Detector, Type[Detector], str], warn: bool = True):
        """Add a ``Detector`` to Scrubber

        You can add a detector to a ``Scrubber`` by passing one of three objects to this function:

            1. the uninitalised class to this function, which initialises the class with default settings.
            2. an instance of a ``Detector`` class, where you can initialise it with the settings desired.
            3. a string containing the name of the detector, which again initialises the class with default settings.

        .. code:: pycon

            >>> import scrubadub
            >>> scrubber = scrubadub.Scrubber(detector_list=[])
            >>> scrubber.add_detector(scrubadub.detectors.CreditCardDetector)
            >>> scrubber.add_detector('skype')
            >>> detector = scrubadub.detectors.DateOfBirthDetector(require_context=False)
            >>> scrubber.add_detector(detector)

        :param detector: The ``Detector`` to add to this scrubber.
        :type detector: a Detector class, a Detector instance, or a string with the detector's name
        :param warn: raise a warning if the locale is not supported by the detector.
        :type warn: bool, default True
        """
        if isinstance(detector, type):
            if not issubclass(detector, Detector):
                raise TypeError((
                    '"%(detector)s" is not a subclass of Detector'
                ) % locals())
            self._check_and_add_detector(detector(locale=self._locale), warn=warn)
        elif isinstance(detector, Detector):
            self._check_and_add_detector(detector, warn=warn)
        elif isinstance(detector, str):
            detector_cls = detectors.catalogue.detector_catalogue.get(detector)
            self._check_and_add_detector(detector_cls(locale=self._locale), warn=warn)

    def remove_detector(self, detector: Union[Detector, Type[Detector], str]):
        """Remove a ``Detector`` from a Scrubber

        You can remove a detector from a ``Scrubber`` by passing one of three objects to this function:

            1. the uninitalised class to this function, which removes the initalised detector of the same name.
            2. an instance of a ``Detector`` class, which removes the initalised detector of the same name.
            3. a string containing the name of the detector, which removed the detector of that name.

        .. code:: pycon

            >>> import scrubadub
            >>> scrubber = scrubadub.Scrubber()
            >>> scrubber.remove_detector(scrubadub.detectors.CreditCardDetector)
            >>> scrubber.remove_detector('url')
            >>> detector = scrubadub.detectors.email.EmailDetector()
            >>> scrubber.remove_detector(detector)

        :param detector: The ``Detector`` to remove from this scrubber.
        :type detector: a Detector class, a Detector instance, or a string with the detector's name
        """
        if isinstance(detector, type):
            self._detectors.pop(detector().name)
        elif isinstance(detector, detectors.base.Detector):
            self._detectors.pop(detector.name)
        elif isinstance(detector, str):
            self._detectors.pop(detector)

    def _check_and_add_detector(self, detector: Detector, warn: bool = False):
        """Check the types and add the detector to the scrubber"""
        if not isinstance(detector, Detector):
            raise TypeError((
                'The detector "{}" is not an instance of the '
                'Detector class.'
            ).format(detector))

        name = detector.name
        if hasattr(detector, 'supported_locale'):
            if not detector.supported_locale(self._locale):  # type: ignore
                if warn:
                    warnings.warn("Detector {} does not support the scrubber locale '{}'.".format(name, self._locale))
        if name in self._detectors:
            raise KeyError((
                'can not add Detector "%(name)s" to this Scrubber, this name is already in use. '
                'Try removing it first.'
            ) % locals())
        self._detectors[name] = detector

    def add_post_processor(self, post_processor: Union[PostProcessor, Type[PostProcessor], str], index: int = None):
        """Add a ``PostProcessor`` to a Scrubber

        You can add a post-processor to a ``Scrubber`` by passing one of three objects to this function:

            1. the uninitalised class to this function, which initialises the class with default settings.
            2. an instance of a ``PostProcessor`` class, where you can initialise it with the settings desired.
            3. a string containing the name of the detector, which again initialises the class with default settings.

        .. code:: pycon

            >>> import scrubadub, scrubadub.post_processors
            >>> scrubber = scrubadub.Scrubber()
            >>> scrubber.add_post_processor('filth_replacer')
            >>> scrubber.add_post_processor(scrubadub.post_processors.PrefixSuffixReplacer)

        :param post_processor: The ``PostProcessor`` to remove from this scrubber.
        :type post_processor: a PostProcessor class, a PostProcessor instance, or a string with the post-processor's
            name
        """
        if isinstance(post_processor, type):
            if not issubclass(post_processor, PostProcessor):
                raise TypeError((
                    '"%(post_processor)s" is not a subclass of PostProcessor'
                ) % locals())
            self._check_and_add_post_processor(post_processor(), index=index)
        elif isinstance(post_processor, PostProcessor):
            self._check_and_add_post_processor(post_processor, index=index)
        elif isinstance(post_processor, str):
            if post_processor in post_processors.catalogue.post_processor_catalogue:
                self._check_and_add_post_processor(
                    post_processors.catalogue.post_processor_catalogue.get(post_processor)(), index=index
                )
            else:
                raise ValueError("Unknown PostProcessor: {}".format(post_processor))

    def remove_post_processor(self, post_processor: Union[PostProcessor, Type[PostProcessor], str]):
        """Remove a ``PostProcessor`` from a Scrubber

        You can remove a post-processor from a ``Scrubber`` by passing one of three objects to this function:

            1. the uninitalised class to this function, which removes the initalised post-processor of the same name.
            2. an instance of a ``PostProcessor`` class, which removes the initalised post-processor of the same name.
            3. a string containing the name of the detector, which removed the post-processor of that name.

        .. code:: pycon

            >>> import scrubadub, scrubadub.post_processors
            >>> scrubber = scrubadub.Scrubber()
            >>> scrubber.remove_post_processor('filth_type_replacer')
            >>> scrubber.remove_post_processor(scrubadub.post_processors.PrefixSuffixReplacer)

        :param post_processor: The ``PostProcessor`` to remove from this scrubber.
        :type post_processor: a PostProcessor class, a PostProcessor instance, or a string with the post-processor's
            name
        """
        if isinstance(post_processor, type):
            self._post_processors = [x for x in self._post_processors if x.name != post_processor().name]
        elif isinstance(post_processor, post_processors.base.PostProcessor):
            self._post_processors = [x for x in self._post_processors if x.name != post_processor.name]
        elif isinstance(post_processor, str):
            self._post_processors = [x for x in self._post_processors if x.name != post_processor]

    def _check_and_add_post_processor(self, post_processor: PostProcessor, index: int = None):
        """Check the types and add the PostProcessor to the scrubber"""
        if not isinstance(post_processor, PostProcessor):
            raise TypeError((
                'The PostProcessor "{}" is not an instance of the '
                'PostProcessor class.'
            ).format(post_processor))
        name = post_processor.name
        if name in [pp.name for pp in self._post_processors]:
            raise KeyError((
                'can not add PostProcessor "%(name)s" to this Scrubber, this name is already in use. '
                'Try removing it first.'
            ) % locals())
        if index is None:
            self._post_processors.append(post_processor)
        else:
            self._post_processors.insert(index, post_processor)

    def clean(self, text: str, **kwargs) -> str:
        """This is the master method that cleans all of the filth out of the
        dirty dirty ``text``. All keyword arguments to this function are passed
        through to the  ``Filth.replace_with`` method to fine-tune how the
        ``Filth`` is cleaned.
        """
        if 'replace_with' in kwargs:
            warnings.warn("Use of replace_with is depreciated in favour of using PostProcessors", DeprecationWarning)

        # We are collating all Filths so that they can all be passed to the post processing step together.
        # This is needed for some operations within the PostProcesssors.
        # It could be improved if we know which post processors need collated Filths.
        filth_list = list(self.iter_filth(text, document_name=None))  # type: Sequence[Filth]
        filth_list = self._post_process_filth_list(filth_list)
        return self._replace_text(text=text, filth_list=filth_list, document_name=None, **kwargs)

    def clean_documents(self, documents: Union[Sequence[str], Dict[Optional[str], str]], **kwargs) -> \
            Union[Dict[Optional[str], str], Sequence[str]]:
        """This is the master method that cleans all of the filth out of the
        dirty dirty ``text``. All keyword arguments to this function are passed
        through to the  ``Filth.replace_with`` method to fine-tune how the
        ``Filth`` is cleaned.
        """
        if 'replace_with' in kwargs:
            warnings.warn("Use of replace_with is depreciated in favour of using PostProcessors", DeprecationWarning)

        # We are collating all Filths so that they can all be passed to the post processing step together.
        # This is needed for some operations within the PostProcesssors.
        # It could be improved if we know which post processors need collated Filths.
        filth_list = []  # type: Sequence[Filth]
        if isinstance(documents, (list, dict)):
            filth_list = list(self.iter_filth_documents(documents=documents, run_post_processors=True))
        else:
            raise TypeError(
                'documents type should be one of: list of strings or a dict of strings with the key as the '
                'document title.'
            )

        if isinstance(documents, list):
            clean_documents = [
                self._replace_text(text=text, filth_list=filth_list, document_name=str(name), **kwargs)
                for name, text in enumerate(documents)
            ]  # type: Union[Dict[Optional[str], str], Sequence[str]]
        elif isinstance(documents, dict):
            clean_documents = {
                name: self._replace_text(text=text, filth_list=filth_list, document_name=name, **kwargs)
                for name, text in documents.items()
            }

        return clean_documents

    def _replace_text(
            self, text: str, filth_list: Sequence[Filth], document_name: Optional[str], **kwargs
    ) -> str:
        filth_list = [filth for filth in filth_list if filth.document_name == document_name]
        if len(filth_list) == 0:
            return text

        filth_list = self._sort_filths(filth_list)  # TODO: expensive sort may not be needed
        filth = None  # type: Optional[Filth]
        clean_chunks = []
        for next_filth in filth_list:
            clean_chunks.append(text[(0 if filth is None else filth.end):next_filth.beg])
            if next_filth.replacement_string is not None:
                clean_chunks.append(next_filth.replacement_string)
            else:
                clean_chunks.append(next_filth.replace_with(**kwargs))
            filth = next_filth
        if filth is not None:
            clean_chunks.append(text[filth.end:])
        return u''.join(clean_chunks)

    def _post_process_filth_list(self, filth_list: Sequence[Filth]) -> Sequence[Filth]:
        # We are collating all Filths so that they can all be passed to the post processing step together.
        # This is needed for some operations within the PostProcesssors.
        # It could be improved if we know which post processors need collated Filths.
        for post_processor in self._post_processors:
            filth_list = post_processor.process_filth(filth_list)

        return filth_list

    def iter_filth(
            self, text: str, document_name: Optional[str] = None, run_post_processors: bool = True
    ) -> Generator[Filth, None, None]:
        """Iterate over the different types of filth that can exist.
        """
        # Iterates using iter_filth documents.
        # If a name is not provided, passes a list with one element, [text]

        yield from self.iter_filth_documents(documents={document_name: text},
                                             run_post_processors=run_post_processors)

    @staticmethod
    def _detector_iter_filth_iterator(detector: Detector, document_list: Sequence[str],
                                      document_names: Sequence[Optional[str]]) -> Generator[Filth, None, None]:
        for doc_name, text in zip(document_names, document_list):
            yield from detector.iter_filth(text, document_name=doc_name)

    def iter_filth_documents(
            self,
            documents: Union[Sequence[str], Dict[Optional[str], str]],
            run_post_processors: bool = True
    ) -> Generator[Filth, None, None]:
        """Iterate over the different types of filth that can exist."""
        if not isinstance(documents, (dict, list)):
            raise TypeError('documents must be one of a string, list of strings or dict of strings.')

        # Figures out which detectors have iter_filth_documents and applies to them

        if isinstance(documents, dict):
            document_names = list(documents.keys())
            document_texts = list(documents.values())
        elif isinstance(documents, (tuple, list)):
            document_texts = documents
            document_names = [str(x) for x in range(len(documents))]
        else:
            raise TypeError(f'documents should be one of dict, list or tuple, but got unsupported type: '
                            f'{type(documents)}')

        # currently doing this by aggregating all_filths and then sorting
        # inline instead of with a Filth.__cmp__ method, which is apparently
        # much slower http://stackoverflow.com/a/988728/564709
        #
        # NOTE: we could probably do this in a more efficient way by iterating
        # over all detectors simultaneously. just trying to get something
        # working right now and we can worry about efficiency later
        filth_list = []  # type: List[Filth]
        for name, detector in self._detectors.items():
            try:
                filth_iterator = detector.iter_filth_documents(
                    document_list=document_texts,
                    document_names=document_names,
                )
            except NotImplementedError:
                filth_iterator = self._detector_iter_filth_iterator(
                    detector=detector,
                    document_list=document_texts,
                    document_names=document_names,
                )

            for filth in filth_iterator:
                if not isinstance(filth, Filth):
                    raise TypeError('iter_filth must always yield Filth')
                if not filth.is_valid():
                    continue
                filth_list.append(filth)

        # This is split up so that we only have to use lists if we have to post_process Filth
        if run_post_processors:
            all_filths = list(self._merge_filths(filth_list))
            all_filths = list(self._post_process_filth_list(all_filths))

            # Here we loop over a list of Filth...
            for filth in all_filths:
                yield filth
        else:
            # ... but here, we're using a generator. If we try to use the same variable it would have two types and
            # fail static typing in mypy
            yield from self._merge_filths(filth_list)

    @staticmethod
    def _sort_filths(filth_list: Sequence[Filth]) -> List[Filth]:
        """Sorts a list of filths, needed before merging and concatenating"""
        # Sort by start position. If two filths start in the same place then
        # return the longer one first
        filth_list = list(filth_list)
        filth_list.sort(key=lambda f: (
            str(getattr(f, 'document_name', None) if hasattr(f, 'document_name') else ''), f.beg, -f.end
        ))
        return filth_list

    @staticmethod
    def _merge_filths(filth_list: Sequence[Filth]) -> Generator[Filth, None, None]:
        """This is where the Scrubber does its hard work and merges any
        overlapping filths.
        """
        if not filth_list:
            return

        document_name_set = {f.document_name for f in filth_list}
        document_names = []  # type: Sequence[Optional[str]]
        if None in document_name_set:
            list_with_none = [None]  # type: Sequence[Optional[str]]
            list_with_others = sorted([x for x in document_name_set if x is not None])  # type: Sequence[Optional[str]]
            document_names = list(list_with_none) + list(list_with_others)
        else:
            document_names = sorted([x for x in document_name_set if x is not None])

        for document_name in document_names:
            document_filth_list = Scrubber._sort_filths([f for f in filth_list if f.document_name == document_name])

            filth = document_filth_list[0]
            for next_filth in document_filth_list[1:]:
                if filth.end < next_filth.beg:
                    yield filth
                    filth = next_filth
                else:
                    filth = filth.merge(next_filth)
            yield filth



from typing import Optional, Sequence

from scrubadub.filth import Filth
from scrubadub.post_processors.catalogue import register_post_processor
from scrubadub.post_processors.base import PostProcessor


class PrefixSuffixReplacer(PostProcessor):
    """Add a prefix and/or suffix to the Filth's replacement string.

    >>> import scrubadub
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthReplacer(),
    ... ])
    >>> scrubber.clean("Contact me at 522-368-8530 or hernandezjenna@example.com")
    'Contact me at PHONE or EMAIL'
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthReplacer(),
    ...     scrubadub.post_processors.PrefixSuffixReplacer(prefix='{{', suffix='}}'),
    ... ])
    >>> scrubber.clean("Contact me at 522-368-8530 or hernandezjenna@example.com")
    'Contact me at {{PHONE}} or {{EMAIL}}'
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthReplacer(),
    ...     scrubadub.post_processors.PrefixSuffixReplacer(prefix='<b>', suffix='</b>'),
    ... ])
    >>> scrubber.clean("Contact me at 522-368-8530 or hernandezjenna@example.com")
    'Contact me at <b>PHONE</b> or <b>EMAIL</b>'

    """
    name = 'prefix_suffix_replacer'  # type: str
    autoload = False
    index = 1

    def __init__(self, prefix: Optional[str] = '{{', suffix: Optional[str] = '}}', name: Optional[str] = None):
        super(PrefixSuffixReplacer, self).__init__(name=name)

        self.prefix = prefix
        self.suffix = suffix

    def process_filth(self, filth_list: Sequence[Filth]) -> Sequence[Filth]:
        """Processes the filth to add prefixes and suffixes to the replacement text

        :param filth_list: The text to be hashed
        :type filth_list: Sequence[Filth]
        :return: The processed filths
        :rtype: Sequence[Filth]
        """
        for filth_item in filth_list:
            if filth_item.replacement_string is None:
                filth_item.replacement_string = filth_item.type.upper()

            if self.prefix is not None and self.suffix is not None:
                filth_item.replacement_string = self.prefix + filth_item.replacement_string + self.suffix
            elif self.prefix is not None:
                filth_item.replacement_string = self.prefix + filth_item.replacement_string
            elif self.suffix is not None:
                filth_item.replacement_string = filth_item.replacement_string + self.suffix

        return filth_list


register_post_processor(PrefixSuffixReplacer)

__all__ = ['PrefixSuffixReplacer']



import phonenumbers

from typing import Optional

from scrubadub.detectors.catalogue import register_detector
from .base import Detector
from ..filth import PhoneFilth


@register_detector
class PhoneDetector(Detector):
    """Remove phone numbers from dirty dirty ``text`` using
    `python-phonenumbers <https://github.com/daviddrysdale/python-phonenumbers>`_, a port of a
    Google project to correctly format phone numbers in text.

    Set the locale on the scrubber or detector to set the region used to search for valid phone numbers.
    If the locale is set to 'en_CA' Canadian numbers will be searched for, while setting the local to 'en_GB' searches
    for British numbers.
    """
    filth_cls = PhoneFilth
    name = 'phone'
    autoload = True

    def iter_filth(self, text, document_name: Optional[str] = None):
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """
        # create a copy of text to handle multiple phone numbers correctly
        for match in phonenumbers.PhoneNumberMatcher(text, self.region):
            yield PhoneFilth(
                beg=match.start,
                end=match.end,
                text=match.raw_string,
                detector_name=self.name,
                document_name=document_name,
                locale=self.locale,
            )

    @classmethod
    def supported_locale(cls, locale: str) -> bool:
        """Returns true if this ``Detector`` supports the given locale.

        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str
        :return: ``True`` if the locale is supported, otherwise ``False``
        :rtype: bool
        """
        return True



import re

from typing import Optional, Generator

from scrubadub.detectors.catalogue import register_detector
from .base import RegexDetector
from ..filth import EmailFilth, Filth


@register_detector
class EmailDetector(RegexDetector):
    """Use regular expression magic to remove email addresses from dirty
    dirty ``text``. This method also catches email addresses like ``john at
    gmail.com``.
    """
    filth_cls = EmailFilth
    name = 'email'
    autoload = True

    # there may be better solutions than this out there and this certainly
    # doesn't do that great of a job with people that spell out the
    # hyphenation of their email address, but its a pretty solid start.
    #
    # adapted from https://gist.github.com/dideler/5219706
    regex = re.compile((
        r"\b[a-z0-9!#$%&'*+\/=?^_`{|}~-]"             # start with this character
        r"(?:"
        r"    [\.a-z0-9!#$%&'*+\/=?^_`{|}~-]{0,62}"   # valid next characters (max length 64 chars before @)
        r"    [a-z0-9!#$%&'*+\/=?^_`{|}~-]"           # end with this character
        r")?"
        r"(?:@|\sat\s)"                               # @ or the word 'at' instead
        r"[a-z0-9]"                                   # domain starts like this
        r"(?:"
        r"    (?=[a-z0-9-]*(\.|\sdot\s))"             # A lookahead to ensure there is a dot in the domain
        r"    (?:\.|\sdot\s|[a-z0-9-]){0,251}"        # might have a '.' or the word 'dot' instead
        r"    [a-z0-9]"                               # domain has max 253 chars, ends with one of these
        r")+\b"
    ), re.VERBOSE | re.IGNORECASE)

    at_matcher = re.compile(r"@|\sat\s", re.IGNORECASE)
    dot_matcher = re.compile(r"\.|\sdot\s", re.IGNORECASE)

    def iter_filth(self, text: str, document_name: Optional[str] = None) -> Generator[Filth, None, None]:
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """

        if re.search(self.at_matcher, text) and re.search(self.dot_matcher, text):
            yield from super().iter_filth(text=text, document_name=document_name)



from typing import Optional, Sequence

from ..filth import Filth


class PostProcessor(object):
    name = 'post_processor'  # type: str
    autoload = False  # type: bool
    index = 10000  # type: int

    def __init__(self, name: Optional[str] = None):
        if name is not None:
            self.name = name

    def process_filth(self, filth_list: Sequence[Filth]) -> Sequence[Filth]:
        raise NotImplementedError('must be overridden by base classes')



import re

from scrubadub.detectors.catalogue import register_detector
from .base import RegexDetector
from ..filth import UrlFilth


@register_detector
class UrlDetector(RegexDetector):
    """Use regular expressions to remove URLs that begin with ``http://``,
    ``https://`` or ``www.`` from dirty dirty ``text``.

    With ``keep_domain=True``, this detector only obfuscates the path on a
    URL, not its domain. For example,
    ``http://twitter.com/someone/status/234978haoin`` becomes
    ``http://twitter.com/{{replacement}}``.
    """
    filth_cls = UrlFilth
    name = 'url'
    autoload = True

    # this regular expression is convenient for captures the domain name
    # and the path separately, which is useful for keeping the domain name
    # but sanitizing the path altogether
    regex = re.compile(r'''
        (?P<domain>
            (https?:\/\/(www\.)?|www\.)          # protocol http://, etc
            [\-\w@:%\.\+~\#=]{2,256}\.[a-z]{2,6} # domain name
            /?                                   # can have a trailing slash
        )(?P<path>
            [\-\w@:%\+\.~\#?&/=]*                # rest of path, query, & hash
        )
    ''', re.VERBOSE)



from faker import Faker
import stdnum.us.ssn

from scrubadub.filth.base import Filth


class SocialSecurityNumberFilth(Filth):
    type = 'social_security_number'

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        ssn = ''
        if faker.locales == ['en_US']:
            while not stdnum.us.ssn.is_valid(ssn):
                ssn = faker.ssn()
        return faker.ssn()

    def is_valid(self) -> bool:
        return stdnum.us.ssn.is_valid(''.join(char for char in self.text if char not in '. -'))




import os
import math
import hashlib

from typing import Sequence, Optional, Union, Dict
from collections import defaultdict

from scrubadub.filth import Filth, MergedFilth, TaggedEvaluationFilth
from scrubadub.post_processors.base import PostProcessor
from scrubadub.post_processors.catalogue import register_post_processor
from scrubadub import utils


class FilthReplacer(PostProcessor):
    """Creates tokens that are used to replace the Filth found in the text of a document.

    This can be configured to include the filth type (eg phone, name, email, ...), a unique number for each piece of
    Filth, and a hash of the Filth.

    >>> import scrubadub
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthReplacer(),
    ... ])
    >>> scrubber.clean("Contact me at 522-368-8530 or hernandezjenna@example.com")
    'Contact me at PHONE or EMAIL'
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthReplacer(include_hash=True, hash_salt='example', hash_length=8),
    ... ])
    >>> scrubber.clean("Contact me at 522-368-8530 or hernandezjenna@example.com")
    'Contact me at PHONE-7358BF44 or EMAIL-AC0B8AC3'
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthReplacer(include_count=True),
    ... ])
    >>> scrubber.clean("Contact me at taylordaniel@example.com or hernandezjenna@example.com, "
    ...                "but taylordaniel@example.com is probably better.")
    'Contact me at EMAIL-0 or EMAIL-1, but EMAIL-0 is probably better.'
    """
    name = 'filth_replacer'  # type: str
    autoload = False
    index = 0

    # NOTE: this is not an efficient way to store this in memory. could
    # alternatively hash the type and text and do away with the overhead
    # bits of storing the tuple in the lookup
    typed_lookup = defaultdict(lambda: utils.Lookup(), {})  # type: Dict[str, utils.Lookup]

    def __init__(self, include_type: bool = True, include_count: bool = False, include_hash: bool = False,
                 uppercase: bool = True, separator: Optional[str] = None, hash_length: Optional[int] = None,
                 hash_salt: Optional[Union[str, bytes]] = None, **kwargs):
        """Initialise the FilthReplacer.

        :param include_type:
        :type include_type: bool, default True
        :param include_count:
        :type include_count: bool, default False
        :param include_hash:
        :type include_hash: bool, default False
        :param uppercase: Make the label uppercase
        :type uppercase: bool, default True
        :param separator: Used to separate labels if a merged filth is being replaced
        :type separator: Optional[str], default None
        :param hash_length: The length of the hexadecimal hash
        :type hash_length: Optional[int], default None
        :param hash_salt: The salt used in the hashing process
        :type hash_salt: Optional[Union[str, bytes]], default None
        """
        super(FilthReplacer, self).__init__(**kwargs)
        self.include_type = include_type
        self.include_count = include_count
        self.include_hash = include_hash
        self.uppercase = uppercase
        self.separator = separator or '+'
        self.hash_length = hash_length or 16

        if isinstance(hash_salt, str):
            self.hash_salt = hash_salt.encode('utf8')  # type: bytes
        else:
            self.hash_salt = os.urandom(128)

    @classmethod
    def reset_lookup(cls):
        """Reset the lookups that maintain a map of filth to a numeric ID."""
        cls.typed_lookup = defaultdict(lambda: utils.Lookup(), {})

    def filth_label(self, filth: Filth) -> str:
        """This function takes a filth and creates a label that can be used to replace the original text.

        :param filth: Limit the named entities to those in this list, defaults to ``{'PERSON', 'PER', 'ORG'}``
        :type filth: Filth
        :return: The replacement label that should be used for this `Filth`.
        :rtype: str

        """
        filths = [filth]
        if isinstance(filth, MergedFilth):
            filths = filth.filths

        replacements = set()
        for f in filths:
            replacement_pieces = []

            if self.include_type:
                filth_type = getattr(f, 'type', None)
                if filth_type is None:
                    continue
                if filth_type == TaggedEvaluationFilth.type:
                    filth_comparison_type = getattr(f, 'comparison_type', None)
                    if filth_comparison_type is not None:
                        filth_type += '_' + filth_comparison_type
                filth_type = filth_type.replace(' ', '_')

                replacement_pieces.append(filth_type)

            if self.include_count:
                replacement_pieces.append(str(FilthReplacer.typed_lookup[filth_type][f.text.lower()]))

            if self.include_hash:
                replacement_pieces.append(FilthReplacer.get_hash(f.text.lower(), self.hash_salt, self.hash_length))

            if len(replacement_pieces) == 0:
                replacement_pieces = ['filth']

            replacements.add('-'.join(replacement_pieces))

        label = self.separator.join(sorted(replacements))
        if self.uppercase:
            label = label.upper()
        return label

    @staticmethod
    def get_hash(text: str, salt: bytes, length: int) -> str:
        """Get a hash of some text, that has been salted and truncated.

        :param text: The text to be hashed
        :type text: str
        :param salt: The salt that should be used in this hashing
        :type salt: bytes
        :param length: The number of characters long that the hexadecimal hash should be
        :type length: int
        :return: The hash of the text
        :rtype: str
        """
        return hashlib.pbkdf2_hmac(
            hash_name='sha256',
            password=text.encode('utf8'),
            salt=salt,
            iterations=100000,
            dklen=math.ceil(length / 2),
        ).hex()[:length]

    def process_filth(self, filth_list: Sequence[Filth]) -> Sequence[Filth]:
        """Processes the filth to replace the original text

        :param filth_list: The text to be hashed
        :type filth_list: Sequence[Filth]
        :return: The processed filths
        :rtype: Sequence[Filth]
        """
        for filth_item in filth_list:
            filth_item.replacement_string = self.filth_label(filth=filth_item)

        return filth_list


register_post_processor(FilthReplacer)

__all__ = ['FilthReplacer']



import re
import nltk
import textblob

from textblob.blob import BaseBlob
from textblob.en.taggers import PatternTagger

from typing import Optional, Generator

from scrubadub.detectors.catalogue import register_detector
from .base import RegexDetector
from ..filth import SkypeFilth, Filth

# BaseBlob uses NLTKTagger as a pos_tagger, but it works wrong
BaseBlob.pos_tagger = PatternTagger()


@register_detector
class SkypeDetector(RegexDetector):
    """Skype usernames tend to be used inline in dirty dirty text quite
    often but also appear as ``skype: {{SKYPE}}`` quite a bit. This method
    looks at words within ``word_radius`` words of "skype" for things that
    appear to be misspelled or have punctuation in them as a means to
    identify skype usernames.

    Default ``word_radius`` is 10, corresponding with the rough scale of
    half of a sentence before or after the word "skype" is used. Increasing
    the ``word_radius`` will increase the false positive rate and
    decreasing the ``word_radius`` will increase the false negative rate.
    """
    filth_cls = SkypeFilth
    name = 'skype'
    autoload = False

    word_radius = 10

    # these two regular expressions are used to validate a skype usernames.
    # _TOKEN is the core regular expression that is used to chunk text into
    # tokens to make sure all valid skype usernames are considered the same
    # token. Importantly, the word "skype" must pass the _SKYPE regex.
    # SKYPE_TOKEN is used to tokenize text and SKYPE_USERNAME is the same thing
    # but with the 6-32 character limit imposed on the username. adapted from
    # http://bit.ly/1FQs1hD
    _SKYPE = r'[a-zA-Z][a-zA-Z0-9_\-\,\.]'
    SKYPE_TOKEN = _SKYPE + '+'
    SKYPE_USERNAME = re.compile(_SKYPE+'{5,31}')

    def iter_filth(self, text, document_name: Optional[str] = None) -> Generator[Filth, None, None]:
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """

        # find 'skype' in the text using a customized tokenizer. this makes
        # sure that all valid skype usernames are kept as tokens and not split
        # into different words
        tokenizer = nltk.tokenize.regexp.RegexpTokenizer(
            self.SKYPE_TOKEN
        )
        blob = textblob.TextBlob(text, tokenizer=tokenizer)
        skype_indices, tokens = [], []
        for i, token in enumerate(blob.tokens):
            tokens.append(token)
            if 'skype' in token.lower():
                skype_indices.append(i)

        # go through the words before and after skype words to identify
        # potential skype usernames.
        skype_usernames = []
        for i in skype_indices:
            jmin = max(i-self.word_radius, 0)
            jmax = min(i+self.word_radius+1, len(tokens))
            for j in list(range(jmin, i)) + list(range(i+1, jmax)):
                token = tokens[j]
                if self.SKYPE_USERNAME.match(token):

                    # this token is a valid skype username. Most skype
                    # usernames appear to be misspelled words. Word.spellcheck
                    # does not handle the situation of an all caps word very
                    # well, so we cast these to all lower case before checking
                    # whether the word is misspelled
                    if token.isupper():
                        token = token.lower()
                    word = textblob.Word(token)
                    suggestions = word.spellcheck()
                    corrected_word, score = suggestions[0]
                    if score < 0.5:
                        skype_usernames.append(token)

        # replace all skype usernames
        if skype_usernames:
            self.regex = re.compile('|'.join(skype_usernames))
            yield from super(SkypeDetector, self).iter_filth(text, document_name=document_name)

        return



import re

from .base import RegexDetector
from ..filth import CredentialFilth
from scrubadub.detectors.catalogue import register_detector


@register_detector
class CredentialDetector(RegexDetector):
    """Remove username/password combinations from dirty drity ``text``.
    """
    filth_cls = CredentialFilth
    name = 'credential'
    autoload = True

    # this regular expression searches for patterns like
    #     "username: root password: root"
    # that tend to occur very frequently in text. This does not currently catch
    # things like "username / password is root / root"
    regex = re.compile(r'''
        (username|login|u:)\s*:?\s*    # username might have : and whitespace
        (?P<username>[\w\-\.@+]*)      # capture the username for replacement
        \s+                            # some whitespace between
        (password|pw|p:)\s*:?\s*       # password might have : and whitespace
        (?P<password>.*)               # password can be anything until EOL
    ''', re.MULTILINE | re.VERBOSE | re.IGNORECASE)



from .catalogue import post_processor_catalogue, register_post_processor, remove_post_processor

from .base import PostProcessor
from .filth_replacer import FilthReplacer
from .prefix_suffix import PrefixSuffixReplacer
from .remover import FilthRemover



from .base import Filth, MergedFilth, RegexFilth
from .address import AddressFilth
from .credential import CredentialFilth
from .credit_card import CreditCardFilth
from .drivers_licence import DriversLicenceFilth
from .email import EmailFilth
from .tagged import TaggedEvaluationFilth
from .location import LocationFilth
from .name import NameFilth
from .organization import OrganizationFilth
from .phone import PhoneFilth
from .postalcode import PostalCodeFilth
from .skype import SkypeFilth
from .twitter import TwitterFilth
from .url import UrlFilth
from .vehicle_licence_plate import VehicleLicencePlateFilth
from .date_of_birth import DateOfBirthFilth
from .en_GB.national_insurance_number import NationalInsuranceNumberFilth
from .en_GB.tax_reference_number import TaxReferenceNumberFilth
from .en_US.social_security_number import SocialSecurityNumberFilth



from .base import Filth

import typing


class TaggedEvaluationFilth(Filth):
    type = 'tagged'

    def __init__(self, *args, comparison_type: typing.Optional[str] = None, **kwargs):
        super(TaggedEvaluationFilth, self).__init__(*args, **kwargs)
        self.comparison_type = comparison_type



from .catalogue import remove_detector, register_detector, detector_catalogue

from .base import Detector, RegexDetector, RegionLocalisedRegexDetector
from .credential import CredentialDetector
from .credit_card import CreditCardDetector
from .date_of_birth import DateOfBirthDetector
from .drivers_licence import DriversLicenceDetector
from .email import EmailDetector
from .phone import PhoneDetector
from .postalcode import PostalCodeDetector
from .skype import SkypeDetector
from .tagged import TaggedEvaluationFilthDetector
from .text_blob import TextBlobNameDetector
from .twitter import TwitterDetector
from .url import UrlDetector
from .user_supplied import UserSuppliedFilthDetector
from .vehicle_licence_plate import VehicleLicencePlateDetector

from . import en_GB
from . import en_US



import re
import copy
import random
import itertools

from faker import Faker

from . import filth as filth_module
from .filth import Filth
from .detectors.tagged import KnownFilthItem

from typing import List, Dict, Union, Optional, Tuple, Callable, Iterable, Type, Set
import numpy as np
import pandas as pd
import sklearn.metrics

# I was originally thinking of building this into the Filth system, but they serve subtlly different purposes:
#   * Filths need to be merged by text location so that replacements can be made
#   * TextPostions need to be merged by text location, but seperated by type so that we an correclty count them
from .utils import ToStringMixin

Grouping = Dict[str, str]
GroupingFunction = Callable[[Filth], Grouping]


class TextPosition(ToStringMixin):
    def __init__(self, filth: Filth, grouping_function: GroupingFunction):
        self.beg = filth.beg
        self.end = filth.end
        self.detected = set()  # type: Set[Tuple[str, ...]]
        self.tagged = set()  # type: Set[Tuple[str, ...]]
        self.document_name = str(filth.document_name or '')  # type: str

        if isinstance(filth, filth_module.TaggedEvaluationFilth):
            self.tagged.add(tuple(grouping_function(filth).values()))
        else:
            self.detected.add(tuple(grouping_function(filth).values()))

    @staticmethod
    def sort_key(position: 'TextPosition') -> Tuple[str, int, int]:
        return (position.document_name, position.beg, -position.end)

    def merge(self, other: 'TextPosition') -> 'TextPosition':
        if self.document_name != other.document_name:
            raise ValueError("Positions are in different documents")
        if self.beg <= other.end and self.end > other.beg:
            self.beg = min(self.beg, other.beg)
            self.end = max(self.end, other.end)
            self.tagged = self.tagged | other.tagged
            self.detected = self.detected | other.detected
            return self
        raise ValueError(f"Positions do not overlap {self.beg} to {self.end} and {other.beg} to {other.end}")

    def __repr__(self) -> str:
        return self._to_string(['beg', 'end', 'tagged', 'detected', 'document_name', ])

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FilthTypePositions(ToStringMixin, object):
    def __init__(self, grouping_function: GroupingFunction, filth_type: str):
        self.positions = []  # type: List[TextPosition]
        self.filth_type = filth_type
        self.grouping_function = grouping_function
        self.column_names = None  # type: Optional[List[str]]

    def __repr__(self) -> str:
        return self._to_string(['filth_type', 'positions', ])

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def add_filth(self, filth: Filth):
        self.positions.append(TextPosition(filth, grouping_function=self.grouping_function))
        if self.column_names is None:
            self.column_names = list(self.grouping_function(filth).keys())

    @staticmethod
    def _merge_position_list(position_list: List[TextPosition]) -> List[TextPosition]:
        position_list.sort(key=TextPosition.sort_key)
        merged_positions = []  # type: List[TextPosition]

        current_position = position_list[0]
        for next_position in position_list[1:]:
            if current_position.document_name != next_position.document_name or \
                    current_position.end <= next_position.beg:
                merged_positions.append(current_position)
                current_position = next_position
            else:
                current_position = current_position.merge(next_position)
        merged_positions.append(current_position)

        return merged_positions

    def merge_positions(self):
        self.positions = self._merge_position_list(self.positions)

    def get_counts(self) -> pd.DataFrame:
        self.merge_positions()

        data_list = []  # type: List[Dict[Tuple[str, ...], int]]
        for position in self.positions:
            row = {
                detected_name: 1
                for detected_name in position.detected
            }
            row.update({
                detected_name: 1
                for detected_name in position.tagged
            })
            data_list.append(row)

        dataframe = pd.DataFrame(data_list).fillna(0).astype(int)
        dataframe.columns = pd.MultiIndex.from_tuples(
            dataframe.columns.values.tolist(),
            names=self.column_names,
        )

        return dataframe


class FilthGrouper(ToStringMixin, object):
    def __init__(self, filth_types: Optional[List[str]] = None, grouping_function: Optional[GroupingFunction] = None,
                 combine_detectors: bool = False, groupby_documents: bool = False):
        self.types = {}  # type: Dict[str, FilthTypePositions]
        self.combine_detectors = combine_detectors
        self.groupby_documents = groupby_documents
        self.filth_types = filth_types

        if grouping_function is None:
            if self.combine_detectors and self.groupby_documents:
                self.grouping_function = FilthGrouper.grouping_combined_bydoc  # type: GroupingFunction
            elif self.combine_detectors and not self.groupby_documents:
                self.grouping_function = FilthGrouper.grouping_combined
            elif not self.combine_detectors and self.groupby_documents:
                self.grouping_function = FilthGrouper.grouping_default_bydoc
            else:
                self.grouping_function = FilthGrouper.grouping_default
        else:
            self.grouping_function = grouping_function

    def __repr__(self) -> str:
        return self._to_string(['combine_detectors', 'filth_types', 'types', ])

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @staticmethod
    def grouping_default(filth: Filth) -> Grouping:
        detector_name = (
            (filth.detector_name or 'None')
            if not isinstance(filth, filth_module.TaggedEvaluationFilth) else
            filth_module.TaggedEvaluationFilth.type
        )
        result = {
            'filth': getattr(filth, 'comparison_type', filth.type),
            'detector': detector_name,
            'locale': filth.locale or 'None',
        }
        return result

    @staticmethod
    def grouping_default_bydoc(filth: Filth) -> Grouping:
        detector_name = (
            (filth.detector_name or 'None')
            if not isinstance(filth, filth_module.TaggedEvaluationFilth) else
            filth_module.TaggedEvaluationFilth.type
        )
        result = {
            'filth': getattr(filth, 'comparison_type', filth.type),
            'document_name': filth.document_name or 'None',
            'detector': detector_name,
            'locale': filth.locale or 'None',
        }

        return result

    @staticmethod
    def grouping_combined(filth: Filth) -> Grouping:
        detector_name = (
            'combined'
            if not isinstance(filth, filth_module.TaggedEvaluationFilth) else
            filth_module.TaggedEvaluationFilth.type
        )
        return {
            'filth': getattr(filth, 'comparison_type', filth.type),
            'detector': detector_name,
            'locale': filth.locale or 'None',
        }

    @staticmethod
    def grouping_combined_bydoc(filth: Filth) -> Grouping:
        detector_name = (
            'combined'
            if not isinstance(filth, filth_module.TaggedEvaluationFilth) else
            filth_module.TaggedEvaluationFilth.type
        )
        result = {
            'filth': getattr(filth, 'comparison_type', filth.type),
            'document_name': filth.document_name or 'None',
            'detector': detector_name,
            'locale': filth.locale or 'None',
        }

        return result

    def merge_positions(self):
        for positioniser in self.types.values():
            positioniser.merge_positions()

    def add_filths(self, filth_list: List[Filth]):
        for filth_item in filth_list:
            sub_filths = [filth_item]
            if isinstance(filth_item, filth_module.base.MergedFilth):
                sub_filths = filth_item.filths

            for filth in sub_filths:
                filth_type = (
                    filth.comparison_type if isinstance(filth, filth_module.TaggedEvaluationFilth) else filth.type
                )

                if filth_type is None or (self.filth_types is not None and filth_type not in self.filth_types):
                    continue

                if filth_type not in self.types:
                    self.types[filth_type] = FilthTypePositions(
                        filth_type=filth_type,
                        grouping_function=self.grouping_function
                    )

                self.types[filth_type].add_filth(filth)

    @classmethod
    def from_filth_list(
            cls, filth_list: List[Filth], filth_types: Optional[List[str]] = None,
            combine_detectors: bool = False, groupby_documents: bool = False,
            grouping_function: Optional[GroupingFunction] = None
    ) -> 'FilthGrouper':
        grouper = cls(filth_types=filth_types, combine_detectors=combine_detectors, groupby_documents=groupby_documents,
                      grouping_function=grouping_function)
        grouper.add_filths(filth_list)
        grouper.merge_positions()
        return grouper

    def expand_missing(self, df: pd.DataFrame) -> pd.DataFrame:
        set_list = [set(s) for s in zip(*df.columns.values.tolist())]
        for column in itertools.product(*set_list):
            if column not in df.columns:
                df.loc[:, column] = 0
        return df

    def get_counts(self, expand_missing: bool = False) -> pd.DataFrame:
        if len(self.types) == 0:
            return pd.DataFrame()
        df_list = []  # type: List[pd.DataFrame]
        running_rows = 0
        for positioniser in self.types.values():
            pos_df = positioniser.get_counts()
            if expand_missing:
                pos_df = self.expand_missing(pos_df)
            df_list.append(pos_df)
            df_list[-1].index += running_rows
            running_rows += max(df_list[-1].index) + 1
        return pd.concat(df_list).fillna(0).astype(int)


def get_filth_classification_report(
        filth_list: List[Filth],
        combine_detectors: bool = False,
        groupby_documents: bool = False,
        output_dict: bool = False,
) -> Optional[Union[str, Dict[str, float]]]:
    """Evaluates the performance of detectors using KnownFilth.

    An example of using this is shown below:

    .. code:: pycon

        >>> import scrubadub, scrubadub.comparison, scrubadub.detectors.text_blob
        >>> scrubber = scrubadub.Scrubber(detector_list=[
        ...     scrubadub.detectors.text_blob.TextBlobNameDetector(name='name_detector'),
        ...     scrubadub.detectors.TaggedEvaluationFilthDetector([
        ...         {'match': 'Tom', 'filth_type': 'name'},
        ...         {'match': 'tom@example.com', 'filth_type': 'email'},
        ...     ]),
        ... ])
        >>> filth_list = list(scrubber.iter_filth("Hello I am Tom"))
        >>> print(scrubadub.comparison.get_filth_classification_report(filth_list))
        filth    detector         locale      precision    recall  f1-score   support
        <BLANKLINE>
        name     name_detector    en_US            1.00      1.00      1.00         1
        <BLANKLINE>
                                    accuracy                           1.00         1
                                   macro avg       1.00      1.00      1.00         1
                                weighted avg       1.00      1.00      1.00         1
        <BLANKLINE>

    :param filth_list: The list of detected filth
    :type filth_list: A list of `Filth` objects
    :param combine_detectors: Combine performance of all detectors for the same filth/locale
    :type combine_detectors: bool, optional
    :param groupby_documents: Show performance for each file individually
    :type groupby_documents: bool, optional
    :param output_dict: Return the report in JSON format, defautls to False
    :type output_dict: bool, optional
    :return: The report in JSON (a `dict`) or in plain text
    :rtype: `str` or `dict`
    """
    if len(filth_list) == 0:
        return None

    grouper = FilthGrouper.from_filth_list(filth_list, combine_detectors=combine_detectors,
                                           groupby_documents=groupby_documents)
    results_df = grouper.get_counts(expand_missing=True)

    filth_index = results_df.columns.names.index('filth')
    detector_index = results_df.columns.names.index('detector')
    tagged_column_mask = np.array(
        [x[detector_index] == filth_module.TaggedEvaluationFilth.type for x in results_df.columns]
    )

    # Find filth types that have some tagged filth
    tagged_types = [x[filth_index] for x in results_df.columns[tagged_column_mask]]

    # Select the columns that have some related tagged filth, but are not tagged filth themselves
    detected_columns = [
        x for x in results_df.columns[~tagged_column_mask]
        if x[filth_index] in tagged_types
    ]
    detected_classes = results_df.loc[:, detected_columns].values

    # Take the detected_columns above and find their tagged counterparts
    tagged_columns = [
        (*x[:detector_index], filth_module.TaggedEvaluationFilth.type, *x[detector_index + 1:])
        for x in detected_columns
    ]
    # If they don't have any tagged counterpart, set the column to zero
    for column in tagged_columns:
        if column not in results_df.columns:
            results_df.loc[:, column] = 0
            tagged_column_mask = np.append(tagged_column_mask, [True])

    true_classes = results_df.loc[:, tagged_columns].values

    # Then no true classes were found
    if detected_classes.shape[1] == 0:
        return None

    report_prefix = None  # type: Optional[str]
    if not output_dict:
        report_prefix = ''
        class_labels = [''] * len(detected_columns)
        for i, name in enumerate(results_df.columns.names):
            max_length = max([len(str(columns[i])) for columns in results_df.columns] + [len(name)]) + 4
            class_labels = [
                name + columns[i].ljust(max_length)
                for name, columns in zip(class_labels, detected_columns)
            ]
            report_prefix += name.ljust(max_length)
        class_labels = [
            name
            for name in class_labels
        ]
        if report_prefix is not None:
            report_prefix += '  '
    else:
        base_name = ("{}:" * len(results_df.columns.names)).rstrip(':')
        class_labels = [base_name.format(*x) for x in detected_columns]

    # If there is only one label reshape the data so that
    # the classification_report interprets it less ambiguously
    report_labels = []  # type: List[int]
    if detected_classes.shape[1] == 1:
        detected_classes = detected_classes.T[0]
        true_classes = true_classes.T[0]
        report_labels = [1]
    else:
        report_labels = [class_labels.index(x) for x in sorted(class_labels)]
        class_labels = sorted(class_labels)

    report = sklearn.metrics.classification_report(
        true_classes,
        detected_classes,
        output_dict=output_dict,
        zero_division=0,
        target_names=class_labels,
        labels=report_labels,
        # **extra_args
    )

    if report_prefix is not None:
        report = report_prefix + report.lstrip(' ')

    return report


def get_filth_dataframe(filth_list: List[Filth]) -> pd.DataFrame:
    """Produces a pandas `DataFrame` to allow debugging and improving detectors.

    An example of using this is shown below:

    .. code:: pycon

        >>> import scrubadub, scrubadub.comparison, scrubadub.detectors.text_blob
        >>> scrubber = scrubadub.Scrubber(detector_list=[
        ...     scrubadub.detectors.text_blob.TextBlobNameDetector(name='name_detector'),
        ...     scrubadub.detectors.TaggedEvaluationFilthDetector([
        ...         {'match': 'Tom', 'filth_type': 'name'},
        ...         {'match': 'tom@example.com', 'filth_type': 'email'},
        ...     ]),
        ... ])
        >>> filth_list = list(scrubber.iter_filth("Hello I am Tom"))
        >>> with pd.option_context("display.max_columns", 20):
        ...     print(scrubadub.comparison.get_filth_dataframe(filth_list))  # doctest: +NORMALIZE_WHITESPACE
           group_id  filth_id filth_type  detector_name document_name text  beg  end  \\
        0         0         0       name  name_detector          None  Tom   11   14
        <BLANKLINE>
          locale  known_filth comparison_type known_text  known_beg  known_end  \\
        0  en_US         True             NaN        Tom         11         14
        <BLANKLINE>
          known_comparison_type  exact_match  partial_match  true_positive  \\
        0                  name         True           True           True
        <BLANKLINE>
           false_positive  false_negative
        0           False           False

    :param filth_list: The list of detected filth
    :type filth_list: A list of `Filth` objects
    :return: A `pd.DataFrame` containing infomatoin about the detected `Filth`
    :rtype: `pd.DataFrame`

    """
    results = []
    for group_id, filth_item in enumerate(filth_list):
        sub_filths = [filth_item]
        if isinstance(filth_item, filth_module.base.MergedFilth):
            sub_filths = filth_item.filths
        for filth_id, sub_filth in enumerate(sub_filths):
            results.append({
                'group_id': group_id,
                'filth_id': filth_id,
                'filth_type': sub_filth.type,
                'detector_name': getattr(sub_filth, 'detector_name', float('nan')),
                'document_name': getattr(sub_filth, 'document_name', float('nan')),
                'text': sub_filth.text,
                'beg': sub_filth.beg,
                'end': sub_filth.end,
                'locale': sub_filth.locale,
                'known_filth': isinstance(sub_filth, filth_module.TaggedEvaluationFilth),
                'comparison_type': getattr(sub_filth, 'comparison_type', float('nan')),
            })

    results_df = pd.DataFrame(results, columns=['group_id', 'filth_id', 'filth_type', 'detector_name', 'document_name',
                                                'text', 'beg', 'end', 'locale', 'known_filth', 'comparison_type'])
    suffix_label = '_y_suffix'

    return (
        pd.merge(
            results_df.loc[~results_df['known_filth']],
            results_df.loc[results_df['known_filth'], ['group_id', 'text', 'beg', 'end', 'comparison_type']],
            how='outer',
            left_on=('group_id', 'filth_type'),
            right_on=('group_id', 'comparison_type'),
            suffixes=('', suffix_label)
        )
        .rename(columns=lambda x: x if not x.endswith(suffix_label) else 'known_' + x[:-len(suffix_label)])
        .assign(
            known_filth=lambda df: ~pd.isnull(df['known_text']),
            exact_match=lambda df: (df['text'] == df['known_text']).fillna(False),
            partial_match=lambda df: ((df['beg'] < df['known_end']) & (df['end'] > df['known_beg']).fillna(False)),
            true_positive=lambda df: (~pd.isnull(df['known_text'])) & (~pd.isnull(df['text'])),
            false_positive=lambda df: (pd.isnull(df['known_text'])) & (~pd.isnull(df['text'])),
            false_negative=lambda df: (~pd.isnull(df['known_text'])) & (pd.isnull(df['text'])),
        )
    )


def make_fake_document(
        paragraphs: int = 20, locale: str = 'en_US', seed: Optional[int] = None, faker: Optional[Faker] = None,
        filth_types: Optional[List[str]] = None, fake_text_function: Optional[Callable[..., str]] = None,
        additional_filth_types: Optional[Iterable[Type[Filth]]] = None,
) -> Tuple[str, List[KnownFilthItem]]:
    """Creates a fake document containing `Filth` that needs to be removed. Also returns the list of known filth
    items that are needed by the `TaggedEvaluationFilthDetector`\\ .

    An example of using this is shown below:

    .. code:: pycon

        >>> import scrubadub, scrubadub.comparison
        >>> document, known_filth_items = scrubadub.comparison.make_fake_document(paragraphs=1, seed=1)
        >>> scrubber = scrubadub.Scrubber()
        >>> scrubber.add_detector(scrubadub.detectors.TaggedEvaluationFilthDetector(
        ...     known_filth_items=known_filth_items
        ... ))
        >>> filth_list = list(scrubber.iter_filth(document))
        >>> print(scrubadub.comparison.get_filth_classification_report(filth_list))
        filth    detector    locale      precision    recall  f1-score   support
        <BLANKLINE>
        email    email       en_US            1.00      1.00      1.00         2
        url      url         en_US            1.00      1.00      1.00         1
        <BLANKLINE>
                              micro avg       1.00      1.00      1.00         3
                              macro avg       1.00      1.00      1.00         3
                           weighted avg       1.00      1.00      1.00         3
                            samples avg       1.00      1.00      1.00         3
        <BLANKLINE>

    :param paragraphs: The list of detected filth
    :type paragraphs: int
    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH"
    :type locale: str
    :param seed: The random seed used to generate the document
    :type seed: int, optional
    :param faker: A Faker object that is used to generate the text
    :type faker: int
    :param filth_types: A list of the ``Filth.type`` to generate
    :type filth_types: List[str]
    :param fake_text_function: A function that will generate a 1-3 sentances of text
    :type fake_text_function: Callable, optional
    :return: The document and a list of `KnownFilthItem`\\ s
    :rtype: Tuple[str, List[KnownFilthItem]]

    """
    if faker is None:
        faker = Faker(locale=locale)

    if fake_text_function is None:
        fake_text_function = faker.text

    # TODO: register filth types to build up a dict that can be read from, like the detectors
    possible_filth = [
        filth_module.AddressFilth,
        filth_module.EmailFilth,
        filth_module.NameFilth,
        filth_module.PhoneFilth,
        filth_module.PostalCodeFilth,
        filth_module.SocialSecurityNumberFilth,
        filth_module.TwitterFilth,
        filth_module.UrlFilth,
    ]
    if additional_filth_types is not None:
        possible_filth += list(additional_filth_types)

    if filth_types is not None:
        possible_filth = [filth for filth in possible_filth if filth.type in filth_types]

    if seed is not None:
        Faker.seed(seed)
        random.seed(seed)

    doc = ""
    known_items = []  # type: List[KnownFilthItem]
    for _ in range(paragraphs):
        for _ in range(random.randint(1, 10)):
            text = fake_text_function() + " "
            matches = list(re.finditer(r'[\s.]', text))
            position = random.choice(matches)
            chosen_filth = random.choice(possible_filth)
            pii_text = chosen_filth.generate(faker=faker)
            known_items.append({
                'match': copy.copy(pii_text),
                'filth_type': copy.copy(chosen_filth.type),
            })
            separator = position.group()
            if '\n' in pii_text:
                separator = '\n'
            doc += (
                text[:position.start()] +
                (separator if separator != '.' else separator + ' ') +
                pii_text +
                separator +
                text[position.end():]
            )
        doc += "\n\n"
    return (doc.strip(), known_items)



import re
import locale as locale_module

from typing import Optional, Tuple, List

try:
    unicode  # type: ignore  # tell mypy to ignore the fact that this doesnt exist in python3
except NameError:
    basestring = str  # Compatibility for Python 2 and 3


class CanonicalStringSet(set):
    """Just like a set, except it makes sure that all elements are lower case
    strings.
    """

    def _cast_as_lower(self, x):
        if not isinstance(x, basestring):
            raise TypeError('CanonicalStringSet only works with strings')
        return x.lower()

    def __init__(self, *elements):
        super(CanonicalStringSet, self).__init__()
        if elements:
            self.update(*elements)

    def __contains__(self, element):
        return super(CanonicalStringSet, self).__contains__(
            self._cast_as_lower(element)
        )

    def add(self, element):
        return super(CanonicalStringSet, self).add(
            self._cast_as_lower(element)
        )

    def update(self, elements):
        for element in elements:
            self.add(element)

    def remove(self, element):
        return super(CanonicalStringSet, self).remove(
            self._cast_as_lower(element)
        )

    def discard(self, element):
        return super(CanonicalStringSet, self).discard(
            self._cast_as_lower(element)
        )


class Lookup(object):
    """The Lookup object is used to create an in-memory reference table to
    create unique identifiers for ``Filth`` that is encountered.
    """

    def __init__(self):
        self.table = {}

    def __getitem__(self, key):
        try:
            return self.table[key]
        except KeyError:
            self.table[key] = len(self.table)
            return self.table[key]


def locale_transform(locale: str) -> str:
    """Normalise the locale string, e.g. 'fr' -> 'fr_FR'.

    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
    :type locale: str
    :return: The normalised locale string
    :rtype: str
    """
    normalised = locale_module.normalize(locale.lower())
    if normalised not in locale_module.locale_alias.values():
        raise ValueError("Unknown locale '{}', not in locale.locale_alias".format(locale))
    return normalised


def locale_split(locale: str) -> Tuple[Optional[str], Optional[str]]:
    """Split the locale string into the language and region.

    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
    :type locale: str
    :return: The two-letter language code and the two-letter region code in a tuple.
    :rtype: tuple, (str, str)
    """
    locale = locale_transform(locale)

    regex = r'(?P<language>[0-9a-zA-Z]+)(_(?P<region>[0-9a-zA-Z]+))?' \
            r'(\.(?P<charset>[0-9a-zA-Z-]+)(@(?P<charset2>[0-9a-zA-Z]+))?)?'
    match = re.match(regex, locale)
    if match is None:
        raise ValueError('Locale does not match expected format.')

    return match.group('language').lower(), match.group('region').upper()


class ToStringMixin(object):
    def _to_string(self, attributes: List[str]) -> str:
        item_attributes = [
            "{}={}".format(item, getattr(self, item, None).__repr__())
            for item in attributes
            if getattr(self, item, None) is not None
        ]
        return "<{} {}>".format(self.__class__.__name__, " ".join(item_attributes))



from faker import Faker

from .base import Filth


class UrlFilth(Filth):
    type = 'url'

    # This allows you to keep the domain
    keep_domain = False

    # this can be used to customize the output, particularly when
    # keep_domain=True
    url_placeholder = type.upper()

    @property
    def placeholder(self):
        if self.keep_domain:
            return self.match.group('domain') + self.url_placeholder
        return self.url_placeholder

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        return faker.url()



"""
This modules provides a detector for date of births.

The filter_type = contextual is simply looking for any detected dates that are more than 18 years
and date of birth has been mentioned within 2 previous lines.

Another option is to nuke all date of birth that is between 18 and 100 years.

Using 18+ years may not be suitable for all use cases; use with caution.
"""
import re
import logging
from dateparser.search import search_dates
from datetime import datetime

from typing import Optional, List, Generator

from scrubadub.detectors.catalogue import register_detector
from .base import Detector
from ..filth.base import Filth
from ..filth.date_of_birth import DateOfBirthFilth


@register_detector
class DateOfBirthDetector(Detector):
    """This detector aims to detect dates of birth in text.

    First all possible dates are found, then they are filtered to those that would result in people being between
    ``DateOfBirthFilth.min_age_years`` and ``DateOfBirthFilth.max_age_years``, which default to 18 and 100
    respectively.

    If ``require_context`` is True, we search for one of the possible ``context_words`` near the found date. We search
    up to ``context_before`` lines before the date and up to ``context_after`` lines after the date. The context that
    we search for are terms like `'birth'` or `'DoB'` to increase the likelihood that the date is indeed a date of
    birth. The context words can be set using the ``context_words`` parameter, which expects a list of strings.

    >>> import scrubadub, scrubadub.detectors.date_of_birth
    >>> DateOfBirthFilth.min_age_years = 12
    >>> scrubber = scrubadub.Scrubber(detector_list=[
    ...     scrubadub.detectors.date_of_birth.DateOfBirthDetector(),
    ... ])
    >>> scrubber.clean("I was born on 10-Nov-2008.")
    'I was born {{DATE_OF_BIRTH}}.'

    """
    name = 'date_of_birth'
    filth_cls = DateOfBirthFilth
    autoload = False

    context_words_language_map = {
        'en': ['birth', 'born', 'dob', 'd.o.b.'],
        'de': ['geburt', 'geboren', 'geb', 'geb.'],
    }

    def __init__(self, context_before: int = 2, context_after: int = 1, require_context: bool = True,
                 context_words: Optional[List[str]] = None, **kwargs):
        """Initialise the detector.

        :param context_before: The number of lines of context to search before the date
        :type context_before: int
        :param context_after: The number of lines of context to search after the date
        :type context_after: int
        :param require_context: Set to False if your dates of birth are not near words that provide context (such as
            "birth" or "DOB").
        :type require_context: bool
        :param context_words: A list of words that provide context related to dates of birth, such as the following:
            'birth', 'born', 'dob' or 'd.o.b.'.
        :type context_words: bool
        :param name: Overrides the default name of the :class:``Detector``
        :type name: str, optional
        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str, optional
        """
        super(DateOfBirthDetector, self).__init__(**kwargs)

        self.context_before = context_before
        self.context_after = context_after
        self.require_context = require_context

        try:
            self.context_words = self.context_words_language_map[self.language]
        except KeyError:
            raise ValueError("DateOfBirthDetector does not support language {}.".format(self.language))

        if context_words is not None:
            self.context_words = context_words

        self.context_words = [word.lower() for word in self.context_words]

    def iter_filth(self, text: str, document_name: Optional[str] = None) -> Generator[Filth, None, None]:
        """Search ``text`` for ``Filth`` and return a generator of ``Filth`` objects.

        :param text: The dirty text that this Detector should search
        :type text: str
        :param document_name: Name of the document this is being passed to this detector
        :type document_name: Optional[str]
        :return: The found Filth in the text
        :rtype: Generator[Filth]
        """

        # using the dateparser lib - locale can be set here
        try:
            date_picker = search_dates(text, languages=[self.language])
        except RecursionError:
            logger = logging.getLogger("scrubadub.detectors.date_of_birth.DateOfBirthDetector")
            logger.error(f"The document '{document_name}' caused a recursion error in dateparser.")
            raise
        if date_picker is None:
            return None

        lines = text.split('\n')

        for identified_string, identified_date in date_picker:
            # Skip anything that could be a phone number, dates rarely begin with a plus
            suspected_phone_number = str(identified_string).startswith('+')
            if suspected_phone_number:
                continue

            # Skip any dates that fall outside of the configured age range
            years_since_identified_date = datetime.now().year - identified_date.year
            within_age_range = (DateOfBirthFilth.min_age_years <= years_since_identified_date <=
                                DateOfBirthFilth.max_age_years)
            if not within_age_range:
                continue

            # If its desired, search for context, if no context is found skip this identified date
            if self.require_context:
                found_context = False
                # Search line by line for the identified date string (identified_string)
                for i_line, line in enumerate(lines):
                    if identified_string not in line:
                        continue
                    # when you find the identified_string, search for context
                    from_line = max(i_line - self.context_before, 0)
                    to_line = max(i_line + self.context_after + 1, 0)
                    text_context = ' '.join(lines[from_line:to_line]).lower()
                    found_context = any(context_word in text_context for context_word in self.context_words)
                    # If you find any context around any instances of this string, all instance are PII
                    if found_context:
                        break
                # If we didn't find any context, this isnt PII, so skip this date
                if not found_context:
                    continue

            found_dates = re.finditer(re.escape(identified_string), text)

            for instance in found_dates:
                yield DateOfBirthFilth(
                    beg=instance.start(),
                    end=instance.end(),
                    text=instance.group(),
                    detector_name=self.name,
                    document_name=document_name,
                    locale=self.locale,
                )

    @classmethod
    def supported_locale(cls, locale: str) -> bool:
        """Returns true if this ``Detector`` supports the given locale.

        :param locale: The locale of the documents in the format: 2 letter lower-case language code eg "en", "es".
        :type locale: str
        :return: ``True`` if the locale is supported, otherwise ``False``
        :rtype: bool
        """
        language, region = cls.locale_split(locale)
        return language in cls.context_words_language_map.keys()


__all__ = ['DateOfBirthDetector']



import string
import random

from faker import Faker

from .base import Filth


class AddressFilth(Filth):
    type = 'address'

    @staticmethod
    def _randomise_seperators(address: str) -> str:
        target = random.choice(["comma", "newline", "mixed", "spaces", "no_change"])
        if target == "comma":
            return address.replace('\n', ', ')
        elif target == "newline":
            return address.replace(', ', '\n')
        elif target == "spaces":
            return address.replace(', ', ' ').replace('\n', ' ')
        elif target == "mixed":
            address = address.replace(', ', '{{SEP}}').replace('\n', '{{SEP}}')
            while '{{SEP}}' in address:
                this_seporator = random.choice(["comma", "newline", "spaces"])
                if this_seporator == "comma":
                    address = address.replace('{{SEP}}', ', ', 1)
                elif this_seporator == "newline":
                    address = address.replace('{{SEP}}', '\n', 1)
                elif this_seporator == "spaces":
                    address = address.replace('{{SEP}}', ' ', 1)
            return address
        return address

    @staticmethod
    def _randomise_street_number(address: str) -> str:
        target = random.choice(["remove", "add_letter", "no_change", "no_change", "no_change", "no_change"])
        if target == "remove":
            address_split = address.split('\n')
            first_line_split = address_split[0].split(' ')
            try:
                int(first_line_split[0])
            except ValueError:
                return address
            new_first_line = " ".join(first_line_split[1:])
            return "\n".join([new_first_line] + address_split[1:])
        elif target == "add_letter":
            address_split = address.split('\n')
            first_line_split = address_split[0].split(' ')
            try:
                int(first_line_split[0])
            except ValueError:
                return address
            new_number = first_line_split[0] + random.choice(string.ascii_letters)
            new_first_line = " ".join([new_number] + first_line_split[1:])
            return "\n".join([new_first_line] + address_split[1:])
        return address

    @staticmethod
    def _randomise_postcode(address: str) -> str:
        target = random.choice(["remove", "lower", "no_change", "no_change", "no_change"])
        if target == "remove":
            return "\n".join(address.split('\n')[:-1])
        elif target == "lower":
            address_split = address.split('\n')
            return "\n".join(address.split('\n')[:-1] + [address_split[-1].lower()])
        return address

    @staticmethod
    def _randomise_country(address: str) -> str:
        target = random.choice(["country", "upper_country", "no_change", "no_change", "no_change"])
        if "country" in target:
            country = random.choice(['United Kingdom', 'Britain', 'England', 'Scotland', 'Wales', 'Cymru', 'GB'])
            if "upper" in target:
                country = country.upper()
            return address + "\n" + country
        return address

    @staticmethod
    def _randomise_building(address: str, faker: Faker) -> str:
        target = random.choice(["add_building", "no_change", "no_change", "no_change"])
        if target == "add_building":
            if bool(random.getrandbits(1)):
                building = faker.last_name() + " " + random.choice(["Building", "House", "Block"])
            else:
                building = random.choice(["Building", "House", "Block"]) + " " + faker.last_name()
            return building + "\n" + address
        return address

    @staticmethod
    def _randomise_case(address: str) -> str:
        target = random.random()
        if target >= 0.8:
            if target >= 0.9:
                address = address.upper()
            else:
                address = address.lower()
        return address

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        address = faker.address()
        if faker.locales == ['en_GB']:
            address = AddressFilth._randomise_street_number(address)
            address = AddressFilth._randomise_building(address, faker)
            address = AddressFilth._randomise_postcode(address)
        if faker.locales == ['en_GB']:
            address = AddressFilth._randomise_country(address)
        address = AddressFilth._randomise_seperators(address)
        address = AddressFilth._randomise_case(address)

        return address



import warnings
from faker import Faker
from typing import Optional, ClassVar, Pattern, List, Match

from .. import exceptions
from .. import utils


class Filth(object):
    """This is the base class for all ``Filth`` that is detected in dirty dirty
    text.
    """

    # this allows people to customize the output, especially for placeholder
    # text and identifier replacements
    prefix = u'{{'  # type: ClassVar[str]
    suffix = u'}}'  # type: ClassVar[str]

    # the `type` is used when filths are merged to come up with a sane label
    type = 'unknown'  # type: ClassVar[str]

    # the `lookup` is used to keep track of all of the different types of filth
    # that are encountered across all `Filth` types.
    lookup = utils.Lookup()

    # For backwards compatibility, but this is deprecated.
    regex = None  # type: Optional[Pattern[str]]

    def __init__(self, beg: Optional[int] = None, end: Optional[int] = None, text: Optional[str] = None,
                 match: Optional[Match] = None, detector_name: Optional[str] = None,
                 document_name: Optional[str] = None, replacement_string: Optional[str] = None,
                 locale: Optional[str] = None, **kwargs):

        self.beg = 0  # type: int
        self.end = 0  # type: int
        self.text = ''  # type: str
        self.match = None  # type: Optional[Match]

        if match is not None and isinstance(match, Match):
            self.beg = match.start()
            self.end = match.end()
            self.text = match.string[match.start():match.end()]
            self.match = match

        if beg is not None:
            self.beg = beg
        if end is not None:
            self.end = end
        if text is not None:
            self.text = text

        self.detector_name = detector_name  # type: Optional[str]
        self.document_name = document_name  # type: Optional[str]
        self.replacement_string = replacement_string  # type: Optional[str]
        self.locale = locale  # type: Optional[str]

        if self.beg >= self.end:
            raise ValueError(
                f"Creating invalid filth (self.beg >= self.end): {self}"
            )

    @property
    def placeholder(self) -> str:
        return self.type.upper()

    @property
    def identifier(self) -> str:
        # NOTE: this is not an efficient way to store this in memory. could
        # alternatively hash the type and text and do away with the overhead
        # bits of storing the tuple in the lookup
        i = self.lookup[(self.type, self.text.lower())]
        return u'%s-%d' % (self.placeholder, i)

    def replace_with(self, replace_with: str = 'placeholder', **kwargs) -> str:
        warnings.warn(
            "Filth.replace_with() will be removed in favour of using the more general PostProcessors",
            DeprecationWarning
        )
        if self.prefix != '{{' or self.suffix != '}}':
            warnings.warn(
                "Setting prefixes and suffixes with scrubadub.filth.Filth.prefix or scrubadub.filth.Filth.suffix "
                "is depreciated in favour of using the PrefixSuffixReplacer",
                DeprecationWarning
            )

        if replace_with == 'placeholder':
            return self.prefix + self.placeholder + self.suffix
        # elif replace_with == 'surrogate':
        #     raise NotImplementedError
        elif replace_with == 'identifier':
            return self.prefix + self.identifier + self.suffix
        else:
            raise exceptions.InvalidReplaceWith(replace_with)

    def merge(self, other_filth: 'Filth') -> 'MergedFilth':
        return MergedFilth(self, other_filth)

    def __repr__(self) -> str:
        return self._to_string()

    def _to_string(self, attributes: Optional[List[str]] = None) -> str:
        if attributes is None:
            attributes = ['text', 'document_name', 'beg', 'end', 'comparison_type', 'detector_name', 'locale']

        item_attributes = [
            "{}={}".format(item, getattr(self, item, None).__repr__())
            for item in attributes
            if getattr(self, item, None) is not None
        ]
        return "<{} {}>".format(self.__class__.__name__, " ".join(item_attributes))

    def __eq__(self, other) -> bool:
        """Only test equality on a subset of class attributes and some are optional"""
        match = True

        if not hasattr(other, 'beg') or not hasattr(other, 'end') or not hasattr(other, 'text'):
            raise TypeError("Unsupported comparison with a Filth and {}".format(type(other)))

        match &= (self.beg == other.beg)
        match &= (self.end == other.end)
        match &= (self.text == other.text)

        if hasattr(self, 'document_name') or hasattr(other, 'document_name'):
            match &= (self.document_name == other.document_name)
        if hasattr(self, 'detector_name') or hasattr(other, 'detector_name'):
            match &= (self.detector_name == other.detector_name)

        return match

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        raise NotImplementedError("A generate() function has not been implemented for this Filth")

    def is_valid(self) -> bool:
        return True


class MergedFilth(Filth):
    """This class takes care of merging different types of filth"""

    def __init__(self, a_filth: Filth, b_filth: Filth):
        super(MergedFilth, self).__init__(
            beg=a_filth.beg,
            end=a_filth.end,
            text=a_filth.text,
            document_name=a_filth.document_name,
        )
        self.filths = [a_filth]
        self._update_content(b_filth)

    def _update_content(self, other_filth: Filth):
        """this updates the bounds, text and placeholder for the merged
        filth
        """
        if self.end < other_filth.beg or other_filth.end < self.beg:
            raise exceptions.FilthMergeError(
                "a_filth goes from [%s, %s) and b_filth goes from [%s, %s)" % (
                    self.beg, self.end, other_filth.beg, other_filth.end
                ))

        if self.document_name != other_filth.document_name:
            raise exceptions.FilthMergeError(
                "This MergedFilth is in document {}, but the Filth that is being merged is in another document {}"
                "".format(self.document_name.__repr__(), other_filth.document_name.__repr__())
            )

        # get the text over lap correct
        if self.beg < other_filth.beg:
            first = self  # type: Filth
            second = other_filth  # type: Filth
        else:
            second = self
            first = other_filth
        end_offset = second.end - first.end
        if end_offset > 0:
            self.text = first.text + second.text[-end_offset:]

        # update the beg/end strings
        self.beg = min(self.beg, other_filth.beg)
        self.end = max(self.end, other_filth.end)
        if self.end - self.beg != len(self.text):
            raise exceptions.FilthMergeError("text length isn't consistent")

        # update the placeholder
        self.filths.append(other_filth)
        self._placeholder = '+'.join([filth.type for filth in self.filths])

    @property
    def placeholder(self):
        return self._placeholder.upper()

    def merge(self, other_filth: Filth) -> 'MergedFilth':
        """Be smart about merging filth in this case to avoid nesting merged
        filths.
        """
        self._update_content(other_filth)
        return self

    def __repr__(self) -> str:
        return self._to_string(['filths'])


class RegexFilth(Filth):
    def __init__(self, *args, **kwargs):
        warnings.warn("Use of RegexFilth is depreciated, use Filth directly instead.", DeprecationWarning)
        super(RegexFilth, self).__init__(*args, **kwargs)



import re
import sys
import copy

from typing import Optional, List, Generator

from scrubadub.detectors.catalogue import register_detector
from .base import Detector
from ..filth.base import Filth
from ..filth.tagged import TaggedEvaluationFilth

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict

KnownFilthItem = TypedDict(
    'KnownFilthItem',
    {
        'match': str,
        'match_end': str,
        'limit': int,
        'ignore_case': bool,
        'ignore_whitespace': bool,
        'ignore_partial_word_matches': bool,
        'filth_type': str
    },
    total=False,
)


@register_detector
class TaggedEvaluationFilthDetector(Detector):
    """Use this ``Detector`` to find tag filth as true ``Filth``. This is useful when you want evaluate the
    effectiveness of a Detector using Filth that has been selected by a human.

    Results from this detector are used as the "truth" against which the other detectos are compared. This is done in
    ``scrubadub.comparison.get_filth_classification_report`` where the detecton accuracies are calculated.

    An example of how to use this detector is given below:

    >>> import scrubadub, scrubadub.comparison, scrubadub.detectors.text_blob
    >>> scrubber = scrubadub.Scrubber(detector_list=[
    ...     scrubadub.detectors.text_blob.TextBlobNameDetector(name='name_detector'),
    ...     scrubadub.detectors.TaggedEvaluationFilthDetector([
    ...         {'match': 'Tom', 'filth_type': 'name'},
    ...         {'match': 'tom@example.com', 'filth_type': 'email'},
    ...     ]),
    ... ])
    >>> filth_list = list(scrubber.iter_filth("Hello I am Tom"))
    >>> print(scrubadub.comparison.get_filth_classification_report(filth_list))
    filth    detector         locale      precision    recall  f1-score   support
    <BLANKLINE>
    name     name_detector    en_US            1.00      1.00      1.00         1
    <BLANKLINE>
                                accuracy                           1.00         1
                               macro avg       1.00      1.00      1.00         1
                            weighted avg       1.00      1.00      1.00         1
    <BLANKLINE>

    This detector takes a list of dictonaires (reffered to as known filth items). These specify what to look for in
    the text to label as tagged filth. The dictionary should contain the following keys:

        * ``match`` (`str`) - a string value that will be searched for in the text
        * ``filth_type`` (`str`) - a string value that indicates the type of Filth, should be set to ``Filth.name``.
          An example of these could be 'name' or 'phone' for name and phone filths respectively.

    The known filth item dictionary may also optionally contain:

        * ``match_end`` (`str`) - if specified will search for Filth starting with the value of match and ending with
          the value of ``match_end``
        * ``limit`` (`int`) - an integer describing the maximum number of characters between match and match_end,
          defaults to 150
        * ``ignore_case`` (`bool`) - Ignore case when searching for the tagged filth
        * ``ignore_whitespace`` (`bool`) - Ignore whitespace when matching ("asd qwe" can also match "asd\\\\nqwe")
        * ``ignore_partial_word_matches`` (`bool`) - Ignore matches that are only partial words (if you're looking
          for "Eve", this flag ensure it wont match "Evening")

    Examples of this:

        * ``{'match': 'aaa', 'filth_type': 'name'}`` - will search for an exact match to aaa and return it as a
          ``NameFilth``
        * ``{'match': 'aaa', 'match_end': 'zzz', 'filth_type': 'name'}`` - will search for `aaa` followed by up to 150
          characters followed by `zzz`, which would match both `aaabbbzzz` and `aaazzz`.
        * ``{'match': '012345', 'filth_type': 'phone', 'ignore_partial_word_matches': True}`` - will search for an
          exact match to 012345, ignoring any partial matches and return it as a ``PhoneFilth``

    This detector is not enabled by default (since you need to supply a list of known filths) and so you must always
    add it to your scrubber with a ``scrubber.add_detector(detector)`` call or by adding it to the ``detector_list``
    inialising a ``Scrubber``.
    """

    filth_cls = TaggedEvaluationFilth
    name = 'tagged'
    autoload = False

    def __init__(self, known_filth_items: List[KnownFilthItem], **kwargs):
        """Initialise the ``Detector``.

        :param known_filth_items: A list of dictionaries that describe items to be searched for in the dirty text.
            The keys `match` and `filth_type` are required, which give the text to be searched for and the type of
            filth that the `match` string represents.
            See the class docstring for further details of available flags in this dictionary.
        :type known_filth_items: list of dicts
        :param tagged_filth: Whether the filth has been tagged and should be used as truth when calculating filth
            finding accuracies.
        :type tagged_filth: bool, default True
        :param name: Overrides the default name of the :class:``Detector``
        :type name: str, optional
        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str, optional
        """
        super().__init__(**kwargs)

        for item in known_filth_items:
            if 'match' not in item or 'filth_type' not in item:
                raise KeyError("Each known filth item (dict) needs both keys 'match' and 'filth_type'.")
            if not isinstance(item['match'], str):
                raise ValueError("The value of 'match' in each KnownItem should be a string. "
                                 "Current value: " + item['match'].__repr__())
            if not isinstance(item['filth_type'], str):
                raise ValueError("The value of 'filth_type' in each KnownItem should be a string. "
                                 "Current value: " + item['filth_type'].__repr__())
            item['match'] = item['match'].strip()
            item['filth_type'] = item['filth_type'].strip()
            if 'match_end' in item:
                if not isinstance(item['match_end'], str):
                    raise ValueError("The value of 'match_end' in each KnownItem should be a string. "
                                     "Current value: " + item['match_end'].__repr__())
                item['match_end'] = item['match_end'].strip()

            for key in item.keys():
                if key not in ['match', 'match_end', 'limit', 'filth_type', 'ignore_case', 'ignore_whitespace',
                               'ignore_partial_word_matches']:
                    raise KeyError("Unexpected key '{}' in the known filth item.".format(key))

        self._known_filth_items = self.dedup_dicts(known_filth_items)

    @staticmethod
    def dedup_dicts(known_filth_items: List[KnownFilthItem]) -> List[KnownFilthItem]:
        # It would be nicer to do this with a set, but sets and dictionaries dont work well together, plus this way
        # we get to keep the typing info associated to these dicts.
        deduped = []  # type: List[KnownFilthItem]
        for item in known_filth_items:
            if item not in deduped:
                deduped.append(item)

        return deduped

    def create_filth(
            self, start_location: int, end_location: int, text: str, comparison_type: Optional[str],
            detector_name: str, document_name: Optional[str], locale: str
    ) -> Filth:
        return TaggedEvaluationFilth(
            start_location,
            end_location,
            text,
            comparison_type=comparison_type,
            detector_name=detector_name,
            document_name=document_name,
            locale=locale,
        )

    def _find_all(
            self,
            text: str,
            substr: str,
            comparison_type: Optional[str] = None,
            document_name: Optional[str] = None,
            ignore_case: bool = False,
            ignore_whitespace: bool = False,
            ignore_partial_word_matches: bool = False,
    ) -> Generator[Filth, None, None]:
        """Yield filth for each match to substr in text."""

        text_orig = copy.copy(text)
        if ignore_case:
            text = text.lower()
            substr = substr.lower()

        if ignore_whitespace:
            # We change any white space in the original with "\s+" that has to match one or more whitespace chars
            substr = '\\s+'.join([re.escape(token) for token in substr.split()])
        else:
            substr = re.escape(substr)

        if ignore_partial_word_matches:
            substr = f"\\b{substr}\\b"

        matches = re.finditer(substr, text, re.MULTILINE | re.DOTALL)
        for match in matches:
            yield self.create_filth(
                match.span()[0],
                match.span()[1],
                text_orig[match.span()[0]:match.span()[1]],
                comparison_type=comparison_type,
                detector_name=self.name,
                document_name=document_name,
                locale=self.locale,
            )

    def _find_all_between(
            self,
            text: str,
            substr_start: str,
            substr_end: str,
            limit: int = 150,
            comparison_type: Optional[str] = None,
            document_name: Optional[str] = None,
            ignore_case: bool = False,
            ignore_whitespace: bool = False,
            ignore_partial_word_matches: bool = False,
    ) -> Generator[Filth, None, None]:
        """Yield filth for text between (and including)
        substr_start and substr_end, but only if the text
        between the two is less than limit characters.
        """
        text_orig = copy.copy(text)
        if ignore_case:
            text = text.lower()
            substr_start = substr_start.lower()
            substr_end = substr_end.lower()

        if ignore_whitespace:
            # We change any white space in the original with "\s+" that has to match one or more whitespace chars
            substr_start = '\\s+'.join([re.escape(token) for token in substr_start.split()])
            substr_end = '\\s+'.join([re.escape(token) for token in substr_end.split()])
        else:
            substr_start = re.escape(substr_start)
            substr_end = re.escape(substr_end)

        if ignore_partial_word_matches:
            substr_start = f"\\b{substr_start}\\b"
            substr_end = f"\\b{substr_end}\\b"

        matches = re.finditer(f"({substr_start})(.{{0,{limit}}})({substr_end})", text, re.MULTILINE | re.DOTALL)
        for match in matches:
            yield self.create_filth(
                match.span()[0],
                match.span()[1],
                text_orig[match.span()[0]:match.span()[1]],
                comparison_type=comparison_type,
                detector_name=self.name,
                document_name=document_name,
                locale=self.locale,
            )

    def iter_filth(
            self,
            text: str,
            document_name: Optional[str] = None
    ) -> Generator[Filth, None, None]:
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """
        for pii_item in self._known_filth_items:
            # could also implement other types in here too
            ignore_case = pii_item.get('ignore_case', False)
            ignore_whitespace = pii_item.get('ignore_whitespace', False)
            ignore_partial_word_matches = pii_item.get('ignore_partial_word_matches', False)
            if 'match' in pii_item and 'match_end' in pii_item and pii_item['match_end'] is not None \
                    and len(pii_item['match_end']) > 0:
                yield from self._find_all_between(
                        text,
                        pii_item['match'],
                        pii_item['match_end'],
                        limit=int(pii_item.get('limit', 150) or 150),
                        comparison_type=pii_item.get('filth_type', None),
                        document_name=document_name,
                        ignore_case=ignore_case,
                        ignore_whitespace=ignore_whitespace,
                        ignore_partial_word_matches=ignore_partial_word_matches,
                )
            elif 'match' in pii_item:
                yield from self._find_all(
                        text,
                        pii_item['match'],
                        comparison_type=pii_item.get('filth_type', None),
                        document_name=document_name,
                        ignore_case=ignore_case,
                        ignore_whitespace=ignore_whitespace,
                        ignore_partial_word_matches=ignore_partial_word_matches,
                )
            else:
                raise ValueError(
                    "Unknown keys in predefined PII item: "
                    "{}".format(pii_item.keys())
                )



import re

from scrubadub.detectors.catalogue import register_detector
from scrubadub.detectors.base import RegionLocalisedRegexDetector
from scrubadub.filth import TaxReferenceNumberFilth


@register_detector
class TaxReferenceNumberDetector(RegionLocalisedRegexDetector):
    """Use regular expressions to detect the UK PAYE temporary reference number (TRN),
    Simple pattern matching, no checksum solution.
    """

    name = 'tax_reference_number'
    autoload = True
    filth_cls = TaxReferenceNumberFilth
    # this regex is looking for NINO that does not begin with certain letters
    region_regex = {
        'GB': re.compile(r'''\d{2}\s?[a-zA-Z]{1}(?:\s*\d\s*){5}''', re.IGNORECASE),
    }



from typing import Optional

from scrubadub.detectors.catalogue import register_detector
from .. import filth as filth_module
from ..filth.base import Filth
from .tagged import TaggedEvaluationFilthDetector


@register_detector
class UserSuppliedFilthDetector(TaggedEvaluationFilthDetector):
    """Use this ``Detector`` to find some known filth in the text. An example might be if you have a list of employee
    numbers that you wish to remove from a document, as shown below:

    >>> import scrubadub
    >>> scrubber = scrubadub.Scrubber(detector_list=[
    ...     scrubadub.detectors.UserSuppliedFilthDetector([
    ...         {'match': 'Anika', 'filth_type': 'name'},
    ...         {'match': 'Larry', 'filth_type': 'name'},
    ...     ]),
    ... ])
    >>> scrubber.clean("Anika is my favourite employee.")
    '{{NAME}} is my favourite employee.'

    This detector takes a list of dictonaires (reffered to as known filth items). These specify what to look for in
    the text to label as tagged filth. The dictionary should contain the following keys:

        * ``match`` (`str`) - a string value that will be searched for in the text
        * ``filth_type`` (`str`) - a string value that indicates the type of Filth, should be set to ``Filth.name``.
          An example of these could be 'name' or 'phone' for name and phone filths respectively.

    The known filth item dictionary may also optionally contain:

        * ``match_end`` (`str`) - if specified will search for Filth starting with the value of match and ending with
          the value of ``match_end``
        * ``limit`` (`int`) - an integer describing the maximum number of characters between match and match_end,
          defaults to 150
        * ``ignore_case`` (`bool`) - Ignore case when searching for the tagged filth
        * ``ignore_whitespace`` (`bool`) - Ignore whitespace when matching ("asd qwe" can also match "asd\\\\nqwe")
        * ``ignore_partial_word_matches`` (`bool`) - Ignore matches that are only partial words (if you're looking
          for "Eve", this flag ensure it wont match "Evening")

    Examples of this:

        * ``{'match': 'aaa', 'filth_type': 'name'}`` - will search for an exact match to aaa and return it as a
          ``NameFilth``
        * ``{'match': 'aaa', 'match_end': 'zzz', 'filth_type': 'name'}`` - will search for `aaa` followed by up to 150
          characters followed by `zzz`, which would match both `aaabbbzzz` and `aaazzz`.
        * ``{'match': '012345', 'filth_type': 'phone', 'ignore_partial_word_matches': True}`` - will search for an
          exact match to 012345, ignoring any partial matches and return it as a ``PhoneFilth``

    This detector is not enabled by default (since you need to supply a list of known filths) and so you must always
    add it to your scrubber with a ``scrubber.add_detector(detector)`` call or by adding it to the ``detector_list``
    inialising a ``Scrubber``.
    """

    name = 'user_supplied'

    def create_filth(
            self, start_location: int, end_location: int, text: str, comparison_type: Optional[str],
            detector_name: str, document_name: Optional[str], locale: str
    ) -> Filth:
        for item_name in dir(filth_module):
            try:
                filth_cls = filth_module.__getattribute__(item_name)
            except AttributeError:
                continue

            if not isinstance(filth_cls, type) or not issubclass(filth_cls, Filth):
                continue

            try:
                filth_type = filth_cls.type
            except AttributeError:
                continue

            if filth_type != comparison_type:
                continue

            return filth_cls(
                start_location,
                end_location,
                text,
                detector_name=detector_name,
                document_name=document_name,
                locale=locale,
            )
        raise KeyError(f"Unable to find filth '{comparison_type}'")




from typing import Union, List, Dict, Sequence, Optional

# convenient imports
from .scrubbers import Scrubber
from . import filth
from . import detectors
from . import post_processors
from .filth import Filth

__version__ = VERSION = "2.0.0"
__all__ = [
    'Scrubber', 'filth', 'detectors', 'post_processors', 'clean', 'clean_documents', 'list_filth',
    'list_filth_documents',
]


def clean(text: str, locale: Optional[str] = None, **kwargs) -> str:
    """Seaches for ``Filth`` in `text` in a string and replaces it with placeholders.

    .. code:: pycon

        >>> import scrubadub
        >>> scrubadub.clean(u"contact me at joe@example.com")
        'contact me at {{EMAIL}}'

    :param text: The text containing possible PII that needs to be redacted
    :type text: `str`
    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH"
    :type locale: str
    :return: Text with all :class:``Filth`` replaced.
    :rtype: `str`

    """
    scrubber = Scrubber(locale=locale)
    return scrubber.clean(text, **kwargs)


def clean_documents(documents: Union[Sequence[str], Dict[Optional[str], str]], locale: Optional[str] = None, **kwargs
                    ) -> Union[Sequence[str], Dict[Optional[str], str]]:
    """Seaches for ``Filth`` in `documents` and replaces it with placeholders.

    `documents` can be in a dict, in the format of ``{'document_name': 'document'}``, or as a list of strings
    (each a seperate document).
    This can be useful when processing many documents.

    .. code:: pycon

        >>> import scrubadub
        >>> scrubadub.clean_documents({'contact.txt': "contact me at joe@example.com",
        ...     'hello.txt': 'hello world!'})
        {'contact.txt': 'contact me at {{EMAIL}}', 'hello.txt': 'hello world!'}

        >>> scrubadub.clean_documents(["contact me at joe@example.com", 'hello world!'])
        ['contact me at {{EMAIL}}', 'hello world!']

    :param documents: Documents containing possible PII that needs to be redacted in the form of a list of documents
        or a dictonary with the key as the document name and the value as the document text
    :type documents: `list` of `str` objects, `dict` of `str` objects
    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH"
    :type locale: str
    :return: Documents in the same format as input, but with `Filth` redacted
    :rtype: `list` of `str` objects, `dict` of `str` objects; same as input
    """
    scrubber = Scrubber(locale=locale)
    return scrubber.clean_documents(documents, **kwargs)


def list_filth(text: str, locale: Optional[str] = None, **kwargs) -> List[Filth]:
    """Return a list of ``Filth`` that was detected in the string `text`.

    .. code:: pycon

        >>> import scrubadub
        >>> scrubadub.list_filth(u"contact me at joe@example.com")
        [<EmailFilth text='joe@example.com' beg=14 end=29 detector_name='email' locale='en_US'>]

    :param text: The text containing possible PII that needs to be found
    :type text: `str`
    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH"
    :type locale: str
    :return: A list of all the :class:``Filth`` objects that were found
    :rtype: `list` of :class:``Filth`` objects

    """
    scrubber = Scrubber(locale=locale)
    return list(scrubber.iter_filth(text, **kwargs))


def list_filth_documents(documents: Union[List[str], Dict[Optional[str], str]], locale: Optional[str] = None,
                         **kwargs) -> List[Filth]:
    """Return a list of ``Filth`` that was detected in the string `text`.

    `documents` can be in a dict, in the format of ``{'document_name': 'document'}``, or as a list of strings
    (each a seperate document).
    This can be useful when processing many documents.

    .. code:: pycon

        >>> import scrubadub
        >>> scrubadub.list_filth_documents(
        ...     {'contact.txt': "contact me at joe@example.com", 'hello.txt': 'hello world!'}
        ... )
        [<EmailFilth text='joe@example.com' document_name='contact.txt' beg=14 end=29 detector_name='email' \
locale='en_US'>]

        >>> scrubadub.list_filth_documents(["contact me at joe@example.com", 'hello world!'])
        [<EmailFilth text='joe@example.com' document_name='0' beg=14 end=29 detector_name='email' locale='en_US'>]

    :param documents: Documents containing possible PII that needs to be found in the form of a list of documents
        or a dictonary with the key as the document name and the value as the document text
    :type documents: `list` of `str` objects, `dict` of `str` objects
    :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                   underscore and the two letter upper-case country code, eg "en_GB" or "de_CH"
    :type locale: str
    :return: A list of all the :class:``Filth`` objects that were found
    :rtype: `list` of :class:``Filth`` objects

    """
    scrubber = Scrubber(locale=locale)
    return list(scrubber.iter_filth_documents(documents, **kwargs))





# this is the base exception that is thrown by scrubadub to make it
# easy to suppress all Scrubadub exceptions
class ScrubadubException(Exception):

    def __init__(self, *args, **kwargs):
        self.issues_url = 'http://github.com/LeapBeyond/scrubadub/issues'

    def render(self, msg):
        return msg % vars(self)


class UnicodeRequired(ScrubadubException):
    """Scrubadub requires unicode. Throw a useful error to lead users to
    the promised land.
    """

    def __str__(self):
        return self.render((
            'scrubadub works best with unicode.\n'
            'Frustrated by unicode?\n'
            'Yeah, me too.\n'
            'But unicode sandwiches are awesome.\n'
            'http://bit.ly/unipain @nedbat\n'
        ))


class UnexpectedFilth(ScrubadubException):
    pass


class FilthMergeError(ScrubadubException):
    pass


class InvalidReplaceWith(ScrubadubException):

    def __init__(self, replace_with):
        super(InvalidReplaceWith, self).__init__()
        self.replace_with = replace_with

    def __str__(self):
        return self.render((
            'Invalid replace_with parameter %(replace_with)s. Can only use '
            '`placeholder` for the time being. If you have other ideas for '
            'replace_with functionality, please make a suggestion at '
            '%(issues_url)s'
        ))



from .base import Filth
from .. import exceptions


class CredentialFilth(Filth):
    type = 'credential'

    # specify how the username/password are replaced
    username_placeholder = 'USERNAME'
    password_placeholder = 'PASSWORD'

    @property
    def placeholder(self):
        ubeg, uend = self.match.span('username')
        pbeg, pend = self.match.span('password')
        return (
            self.match.string[self.match.start():ubeg] +
            self.prefix + self.username_placeholder + self.suffix +
            self.match.string[uend:pbeg] +
            self.prefix + self.password_placeholder + self.suffix
        )

    # override the replace_with method for credentials because the
    # prefix/suffix components are mixed into the placeholder
    def replace_with(self, replace_with='placeholder', **kwargs):
        if replace_with == 'placeholder':
            return self.placeholder
        else:
            raise exceptions.InvalidReplaceWith(replace_with)



import re
import textblob

from textblob.blob import BaseBlob
from textblob.en.taggers import PatternTagger

from typing import Optional, Generator

from scrubadub.detectors.catalogue import register_detector
from .base import RegexDetector
from ..filth import NameFilth, Filth
from ..utils import CanonicalStringSet

# BaseBlob uses NLTKTagger as a pos_tagger, but it works wrong
BaseBlob.pos_tagger = PatternTagger()


@register_detector
class TextBlobNameDetector(RegexDetector):
    """Use part of speech tagging from textblob to clean proper nouns out of the dirty dirty
    ``text``. Disallow particular nouns by adding them to the ``NameDetector.disallowed_nouns`` set.
    """
    filth_cls = NameFilth
    name = 'text_blob_name'
    autoload = False

    disallowed_nouns = CanonicalStringSet(["skype"])

    def iter_filth(self, text, document_name: Optional[str] = None) -> Generator[Filth, None, None]:
        """Yields discovered filth in the provided ``text``.

        :param text: The dirty text to clean.
        :type text: str
        :param document_name: The name of the document to clean.
        :type document_name: str, optional
        :return: An iterator to the discovered :class:`Filth`
        :rtype: Iterator[:class:`Filth`]
        """

        if not isinstance(self.disallowed_nouns, CanonicalStringSet):
            raise TypeError(
                'NameDetector.disallowed_nouns must be CanonicalStringSet'
            )

        # find the set of proper nouns using textblob.
        proper_nouns = set()
        blob = textblob.TextBlob(text)
        for word, part_of_speech in blob.tags:
            is_proper_noun = part_of_speech in ("NNP", "NNPS")
            if is_proper_noun and word.lower() not in self.disallowed_nouns:
                proper_nouns.add(word)

        # use a regex to replace the proper nouns by first escaping any
        # lingering punctuation in the regex
        # http://stackoverflow.com/a/4202559/564709
        if proper_nouns:
            re_list = []
            for proper_noun in proper_nouns:
                re_list.append(r'\b' + re.escape(str(proper_noun)) + r'\b')
            self.regex = re.compile('|'.join(re_list))
            yield from super(TextBlobNameDetector, self).iter_filth(text, document_name=document_name)
        return

    @classmethod
    def supported_locale(cls, locale: str) -> bool:
        """Returns true if this ``Detector`` supports the given locale.

        :param locale: The locale of the documents in the format: 2 letter lower-case language code followed by an
                       underscore and the two letter upper-case country code, eg "en_GB" or "de_CH".
        :type locale: str
        :return: ``True`` if the locale is supported, otherwise ``False``
        :rtype: bool
        """
        language, region = cls.locale_split(locale)

        # fr and de are possible through plugins, but need to be implemented on this end
        # https://github.com/sloria/textblob-fr and https://github.com/markuskiller/textblob-de
        return language in ['en', ]



import re

from scrubadub.detectors.catalogue import register_detector
from scrubadub.detectors.base import RegionLocalisedRegexDetector
from scrubadub.filth import SocialSecurityNumberFilth


@register_detector
class SocialSecurityNumberDetector(RegionLocalisedRegexDetector):
    """Use regular expressions to detect a social security number (SSN) in
    dirty dirty ``text``.
    """

    filth_cls = SocialSecurityNumberFilth
    name = 'social_security_number'
    autoload = True
    region_regex = {
        'US': re.compile((
            r"[0-9][0-9][0-9]"  # first three digits
            r"[\-. ]"  # separator
            r"[0-9][0-9]"  # next two digits
            r"[\-. ]"  # separator
            r"[0-9][0-9][0-9][0-9]"  # last four digits
        ), re.VERBOSE),
    }



import re

from scrubadub.detectors.catalogue import register_detector
from .base import RegionLocalisedRegexDetector
from ..filth import DriversLicenceFilth


@register_detector
class DriversLicenceDetector(RegionLocalisedRegexDetector):
    """Use regular expressions to detect UK driving licence numbers,
    Simple pattern matching, no checksum solution.
    """

    name = 'drivers_licence'
    autoload = True
    filth_cls = DriversLicenceFilth

    region_regex = {
        # this regex is looking for UK driving licence numbers that follow a pattern, no checksum
        'GB': re.compile(r'''([a-zA-Z9]{5}\s?)((?:\s*\d\s*){6}[a-zA-Z9]{2}\w{3})\s?(\d{2})''', re.IGNORECASE)
    }



from faker import Faker

from .base import Filth


class PostalCodeFilth(Filth):
    type = "postalcode"

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        # for en_US I expect we should pick between .zipcode() and .zipcode_plus4()
        # as postcode() for en_US only returns the 5 number zip code
        return faker.postcode()



import re

from .base import RegexDetector
from ..filth import CreditCardFilth
from scrubadub.detectors.catalogue import register_detector


@register_detector
class CreditCardDetector(RegexDetector):
    """Remove credit-card numbers from dirty dirty ``text``.

    Supports Visa, MasterCard, American Express, Diners Club and JCB.
    """
    name = 'credit_card'
    filth_cls = CreditCardFilth
    autoload = True

    # Regexes from:
    # http://www.regular-expressions.info/creditcard.html

    # Fake card numbers from:
    # https://www.paypalobjects.com/en_US/vhelp/paypalmanager_help/credit_card_numbers.htm

    # taken from the alphagov fork of scrubadub: https://github.com/alphagov/scrubadub

    # Looking at wikipedia, there are probably more numbers to detect:
    # https://en.wikipedia.org/wiki/Payment_card_number#Issuer_identification_number_.28IIN.29

    # TODO: regex doesn't match if the credit card number has spaces/dashes in

    regex = re.compile((
        r"(?<=\s)"
        r"(?:4[0-9]{12}(?:[0-9]{3})?"  		# Visa
        r"|(?:5[1-5][0-9]{2}"          		# MasterCard
        r"|222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][0-9]|2720)[0-9]{12}"
        r"|3[47][0-9]{13}"             		# American Express
        r"|3(?:0[0-5]|[68][0-9])[0-9]{11}"   	# Diners Club
        r"|6(?:011|5[0-9]{2})[0-9]{12}"      	# Discover
        r"|(?:2131|1800|35\d{3})\d{11})"      	# JCB
    ), re.VERBOSE)



import re

from scrubadub.detectors.catalogue import register_detector
from .base import RegionLocalisedRegexDetector
from ..filth.vehicle_licence_plate import VehicleLicencePlateFilth


@register_detector
class VehicleLicencePlateDetector(RegionLocalisedRegexDetector):
    """Detects standard british licence plates."""
    filth_cls = VehicleLicencePlateFilth
    name = 'vehicle_licence_plate'
    autoload = True

    # Vehicle Registration Plates from:
    # https://gist.github.com/harry-jones/755501192139820eeb65e030fe878f75
    # More cases available in above link, but can cause the regex to become
    # quire greedy. For now keep it simple!

    # taken from the alphagov fork of scrubadub: https://github.com/alphagov/scrubadub

    region_regex = {
        'GB': re.compile(
            # Current system followed by the old system
            r"""
                \b(
                    ([a-zA-Z]{2}[0-9]{2}(?:\s)?[a-zA-Z]{3})
                    |
                    ([a-zA-Z][0-9]{1,3}(?:\s)?[a-zA-Z]{3})
                )\b
            """,
            re.VERBOSE | re.IGNORECASE,
        ),
    }



import random
import datetime
import dateparser
from faker import Faker

from .base import Filth


class DateOfBirthFilth(Filth):
    type = 'date_of_birth'
    min_age_years = 18
    max_age_years = 100

    @staticmethod
    def generate(faker: Faker) -> str:
        """Generates an example of this ``Filth`` type, usually using the faker python library.

        :param faker: The ``Faker`` class from the ``faker`` library
        :type faker: Faker
        :return: An example of this ``Filth``
        :rtype: str
        """
        formats = [
            '%c',  # Tue Aug 16 21:30:00 1988 (en_US); locale dependant
            '%x',  # 08/16/1988 (en_US); locale dependant
            '%a %d %b %Y',  # Sun 19 Jan 1999
            '%A %d %B %Y',  # Sunday 19 January 1999
            '%d-%m-%Y',  # 15-01-1999
            '%A %dth, %B, %Y',  # Monday 08th, January, 1973
        ]
        return faker.date_of_birth().strftime(random.choice(formats))

    def is_valid(self) -> bool:
        """Check to see if the found filth is valid."""
        found_date = dateparser.parse(self.text)
        if found_date is None:
            return False
        years_since_identified_date = datetime.date.today().year - found_date.year
        return DateOfBirthFilth.min_age_years <= years_since_identified_date <= DateOfBirthFilth.max_age_years



from typing import Sequence

from scrubadub.post_processors.catalogue import register_post_processor
from scrubadub.filth import Filth
from scrubadub.post_processors.base import PostProcessor


class FilthRemover(PostProcessor):
    """Removes all found filth from the original document.

    >>> import scrubadub
    >>> scrubber = scrubadub.Scrubber(post_processor_list=[
    ...     scrubadub.post_processors.FilthRemover(),
    ... ])
    >>> scrubber.clean("Contact me at 522-368-8530 or hernandezjenna@example.com")
    'Contact me at  or '

    """
    name = 'filth_remover'  # type: str
    autoload = False
    index = 0

    def process_filth(self, filth_list: Sequence[Filth]) -> Sequence[Filth]:
        """Processes the filth to remove the filth

        :param filth_list: The text to be hashed
        :type filth_list: Sequence[Filth]
        :return: The processed filths
        :rtype: Sequence[Filth]
        """
        for filth_item in filth_list:
            filth_item.replacement_string = ''
        return filth_list


register_post_processor(FilthRemover)

__all__ = ['FilthRemover']



import glob
import os
from setuptools import setup, find_packages

# read in the description from README
with open("README.rst") as stream:
    long_description = stream.read()

github_url = 'https://github.com/LeapBeyond/scrubadub'


def read_packages_from_file(filename):
    with open(filename, 'r') as stream:
        for line in stream:
            package = line.strip().split('#')[0]
            if package:
                yield package


def get_package_list(location):
    location = os.path.join('requirements', location)
    return list(read_packages_from_file(location))


# get the version
version = None
with open(os.path.join('scrubadub', '__init__.py')) as stream:
    for line in stream:
        if 'version' in line.lower():
            version = line.split()[-1].replace('"', '').replace("'", '')

setup(
    name='scrubadub',
    version=version,
    description=(
        "Clean personally identifiable information from dirty dirty text."
    ),
    long_description=long_description,
    url=github_url,
    download_url="%s/archives/master" % github_url,
    author='Dean Malmgren',
    author_email='dean.malmgren@datascopeanalytics.com',
    license='MIT',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing',
        'Topic :: Utilities',
    ],
    install_requires=get_package_list('python'),
    include_package_data=True,
    package_data={'': ['scrubadub/detectors/models/sklearn_address/*.json']},
    zip_safe=False,
)



import inspect
import catalogue

from typing import Type, Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from scrubadub.detectors import Detector

detector_catalogue = catalogue.create('scrubadub', 'detectors', entry_points=True)


def register_detector(detector: Type['Detector'], *, autoload: Optional[bool] = None) -> Type['Detector']:
    """Register a detector for use with the ``Scrubber`` class.

    You can use ``register_detector(NewDetector, autoload=True)`` after your detector definition to automatically
    register it with the ``Scrubber`` class so that it can be used to remove Filth.

    The argument ``autoload``decides whether a new ``Scrubber()`` instance should load this ``detector`` by default.

    .. code:: pycon

        >>> import scrubadub
        >>> class NewDetector(scrubadub.detectors.Detector):
        ...     pass
        >>> scrubadub.detectors.register_detector(NewDetector, autoload=False)
        <class 'scrubadub.detectors.catalogue.NewDetector'>

    :param detector: The ``Detector`` to register with the scrubadub detector configuration.
    :type detector: Detector class
    :param autoload: Whether to automatically load this ``Detector`` on ``Scrubber`` initialisation.
    :type autoload: Optional[bool]
    """
    if not inspect.isclass(detector):
        raise ValueError("detector should be a class, not an instance.")

    if autoload is not None:
        detector.autoload = autoload

    detector_catalogue.register(detector.name, func=detector)

    return detector


def remove_detector(detector: Union[Type['Detector'], str]):
    """Remove an already registered detector.

    .. code:: pycon

        >>> import scrubadub
        >>> class NewDetector(scrubadub.detectors.Detector):
        ...     pass
        >>> scrubadub.detectors.catalogue.register_detector(NewDetector, autoload=False)
        <class 'scrubadub.detectors.catalogue.NewDetector'>
        >>> scrubadub.detectors.catalogue.remove_detector(NewDetector)

    :param detector: The ``Detector`` to register with the scrubadub detector configuration.
    :type detector: Union[Type['PostProcessor'], str]
    :param autoload: Whether to automatically load this ``Detector`` on ``Scrubber`` initialisation.
    :type autoload: bool
    """
    if isinstance(detector, str):
        if detector in detector_catalogue:
            catalogue._remove((*detector_catalogue.namespace, detector))

    elif inspect.isclass(detector):
        if detector.name in detector_catalogue:
            catalogue._remove((*detector_catalogue.namespace, detector.name))

    else:
        raise ValueError("detector should be a class (not an instance) or a string.")
